// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001 Universitaet Karlsruhe, Germany
//                    and Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General
// Public License. See LICENSE.TXT for details.
//





package de.uka.ilkd.key.java;

import de.uka.ilkd.key.java.declaration.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.java.reference.*;
import de.uka.ilkd.key.java.recoderext.*;
import de.uka.ilkd.key.java.expression.literal.NullLiteral;
import de.uka.ilkd.key.java.abstraction.*;
import de.uka.ilkd.key.logic.*;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;
import java.util.Collection;
import java.util.Iterator;
import de.uka.ilkd.key.util.Debug;
import recoder.list.*;
import java.util.Vector;
import de.uka.ilkd.key.java.recoderext.KeYCrossReferenceServiceConfiguration;

public class KeYProgModelInfo{


    private recoder.CrossReferenceServiceConfiguration sc = null;
    private KeYRecoderMapping mapping;
    private TypeConverter typeConverter;

    public KeYProgModelInfo(recoder.CrossReferenceServiceConfiguration crsc,
			    KeYRecoderMapping krm,
			    TypeConverter typeConverter){
        sc = crsc;
	this.typeConverter=typeConverter;
	this.mapping=krm;
    }

    public KeYProgModelInfo(KeYRecoderMapping krm,
			    TypeConverter typeConverter){
	this(new KeYCrossReferenceServiceConfiguration(), krm, 
	     typeConverter);
    }

    public KeYRecoderMapping rec2key() {
	return mapping;
    }

    /**
     * Returns all KeY-elements mapped by Recoder2KeY object of this
     * KeYProgModelInfo.
     * @return a Set object containing the KeY-elements.
     */

    public Set allElements(){
        return rec2key().elemsKeY();
    }


    /**
     * Returns all ObjectSorts mapped to java.Types in KeY.
     * @return a Collection containing the ObjectSorts.
     */
    public Collection allObjectSorts(){
	Set result=new HashSet();
	Iterator it=allElements().iterator();
	while (it.hasNext()) {
	    Object o=it.next();
	    if (o instanceof KeYJavaType && 
		(((KeYJavaType)o).getSort() instanceof ObjectSort));
	    result.add(((KeYJavaType)o).getSort());
	}
        return result;
    }


    private recoder.list.MethodList getAllRecoderMethods(TypeDeclaration td){
        recoder.java.declaration.TypeDeclaration rtd
            = (recoder.java.declaration.TypeDeclaration)rec2key().toRecoder(td);
        return rtd.getAllMethods();
    }


    /**Returns all visible methods that are defined in this
     * class type or any of its supertypes. The methods are
     * in topological order with respect to the inheritance hierarchy.
     * @return the list of visible methods of this type and its supertypes.
     */

    public ListOfMethod getAllMethods(TypeDeclaration td) {
        recoder.list.MethodList rmethods=getAllRecoderMethods(td);
        ListOfMethod result = SLListOfMethod.EMPTY_LIST;
        for (int i=rmethods.size()-1; i>=0; i--) {
            recoder.abstraction.Method rm=rmethods.getMethod(i);
            Method m=(Method)rec2key().toKeY(rm);
            result=result.prepend(m);
        }
        return result;
    }


    /**Returns all visible methods that are defined in this
     * class type or any of its supertypes. The methods are
     * in topological order with respect to the inheritance hierarchy.
     * @return the list of visible methods of this type and its supertypes.
     */

    public ListOfProgramMethod getAllProgramMethods(TypeDeclaration td) {
        recoder.list.MethodList rmethods=getAllRecoderMethods(td);
        ListOfProgramMethod result = SLListOfProgramMethod.EMPTY_LIST;
        for (int i=rmethods.size()-1; i>=0; i--) {
            recoder.abstraction.Method rm=rmethods.getMethod(i);
            ProgramMethod m=(ProgramMethod)rec2key().toKeY(rm);
            result=result.prepend(m);
        }
        return result;
    }


    private recoder.CrossReferenceServiceConfiguration getServConf() {
	return sc;
    }

    private recoder.list.TypeList getRecoderTypes(ListOfType types) {
        if (types==null) {
            return null;
        }
        IteratorOfType it=types.iterator();
        recoder.list.TypeArrayList tl
            = new recoder.list.TypeArrayList(types.size());
        while (it.hasNext()) {
            tl.add( (recoder.abstraction.Type) rec2key().toRecoder(it.next()));
        }
        return tl;
    }



    /**
     * Returns the full name of a KeYJavaType t.
     * @return the full name of t as a String.
     */

    public String getFullName(KeYJavaType t) {
        recoder.abstraction.Type rt
            = (recoder.abstraction.Type) rec2key().toRecoder(t);
        return rt.getFullName();
    }

    private recoder.abstraction.Type getType(TypeReference tr) {
        recoder.abstraction.Type result;
        if (tr instanceof TypeRef) {
            KeYJavaType kjt=((TypeRef)tr).getKeYJavaType();
            result = (recoder.abstraction.Type)
            rec2key().toRecoder(((TypeRef)tr).getKeYJavaType());
            return result;
        }
        result=getServConf().getSourceInfo().getType
            ((recoder.java.reference.TypeReference)rec2key().toRecoder(tr));
        return result;
    }



    /**
     * Checks whether subType is a subtype of superType or not.
     * @returns true if subType is subtype of superType,
     * false in the other case.
     */

    public boolean isSubtype(KeYJavaType subType, KeYJavaType superType) {
        recoder.abstraction.ClassType rSuperType =
            (recoder.abstraction.ClassType)rec2key().toRecoder(superType);
        recoder.abstraction.ClassType rSubType =
            (recoder.abstraction.ClassType)rec2key().toRecoder(subType);
        return isSubtype(rSubType, rSuperType);
    }



    private boolean isSubtype(recoder.abstraction.ClassType classSubType,
                     recoder.abstraction.ClassType classType) {
        boolean isSub =  getServConf().getSourceInfo().
            isSubtype(classSubType, classType);
        if (!isSub) {
            boolean result= getServConf().getByteCodeInfo().
            isSubtype(classSubType, classType);
            return result;
        } else {
            return true;
        }
    }

    private recoder.list.MethodList getRecoderMethods(KeYJavaType ct){
        recoder.abstraction.ClassType rct
            = (recoder.abstraction.ClassType) rec2key().toRecoder(ct);
        return rct.getProgramModelInfo().getMethods(rct);
    }



    private recoder.list.MethodList getRecoderMethods(KeYJavaType ct,
    			String m,
                      	ListOfType signature){
        recoder.abstraction.ClassType rct
            = (recoder.abstraction.ClassType) rec2key().toRecoder(ct);
        return rct.getProgramModelInfo().getMethods(rct, m,
                                        getRecoderTypes(signature));
    }


    /**
     * Returns the list of most specific methods with the given
     * name that are defined in the given type or in a supertype
     * where they are visible for the given type, and have a signature
     * that is compatible to the given one. If used to resolve a
     * method call, the result should be defined and unambiguous.
     * @param ct the class type to get methods from.
     * @param m the name of the methods in question.
     * @param signature the statical type signature of a callee.
     */

    public ListOfMethod getMethods(KeYJavaType ct, String m,
                      ListOfType signature){
        recoder.list.MethodList rml = getRecoderMethods(ct, m, signature);
        ListOfMethod result = SLListOfMethod.EMPTY_LIST;
        for (int i=rml.size()-1; i>=0; i--) {
            recoder.abstraction.Method rm=rml.getMethod(i);
            Method method=(Method)rec2key().toKeY(rm);
            result=result.prepend(method);
        }
        return result;
    }


	/**
	 * Returns the methods locally defined within the given
	 * class type. If the type is represented in source code,
	 * the returned list matches the syntactic order.
	 * @param ct a class type.
	 */

    public ListOfMethod getMethods(KeYJavaType ct){
        recoder.list.MethodList rml = getRecoderMethods(ct);
        ListOfMethod result = SLListOfMethod.EMPTY_LIST;
        for (int i=rml.size()-1; i>=0; i--) {
            recoder.abstraction.Method rm=rml.getMethod(i);
            Method m=(Method)rec2key().toKeY(rm);
            result=result.prepend(m);
        }
        return result;
    }



    /**
     * Returns the programmethods with the given name that is defined
     * in the given type or in a supertype where it is visible for the
     * given type, and has a signature that is compatible to the given one.
     * @param ct the class type to get methods from.
     * @param m the name of the methods in question.
     * @param signature the statical type signature of a callee.
     * @return the programmethod, if one is found,
     * null if none or more than one programmethod is found (in this case
     * a debug output is written to console).
     */

    public ProgramMethod getProgramMethod(KeYJavaType ct, String m,
                      ListOfType signature) {
        recoder.list.MethodList methodlist =
            getRecoderMethods(ct,m,signature);
        if (methodlist.size()==1) {
            return (ProgramMethod) rec2key().toKeY(methodlist.getMethod(0));
        }
        if (methodlist.size()==0) {
            Debug.out("javainfo: Program Method not found: ",m);
            return null;
        }
        Debug.assert(false);
        return null;
    }





    /**
     * Returns the fields locally defined within the given class type.
     * If the type is represented in source code, the returned list
     * matches the syntactic order.
     * @param a class type
     * @return the list of field members of the given type.
     */

    public ListOfField getAllFields(KeYJavaType ct){
        recoder.abstraction.ClassType rct
        = (recoder.abstraction.ClassType) rec2key().toRecoder(ct);
        recoder.list.FieldList rfl =
            rct.getProgramModelInfo().getAllFields(rct);
        ListOfField result = SLListOfField.EMPTY_LIST;
        for (int i=rfl.size()-1; i>=0; i--) {
            recoder.abstraction.Field rf = rfl.getField(i);
            Field f = (Field)rec2key().toKeY(rf);
            result = result.prepend(f);
        }
        return result;
    }



    private recoder.list.ClassTypeList getRecoderSubtypes(KeYJavaType ct){
        recoder.abstraction.ClassType rct
        = (recoder.abstraction.ClassType) rec2key().toRecoder(ct);
        return sc.getCrossReferenceSourceInfo().getSubtypes(rct);
    }


    /**
     * Returns all known subtypes of the given class type.
     * @param a class type
     * @return the list of the known direct subtypes of the given class type.
     */

    public ListOfKeYJavaType getSubtypes(KeYJavaType ct){
        recoder.abstraction.ClassType rct
        = (recoder.abstraction.ClassType) rec2key().toRecoder(ct);
        recoder.list.ClassTypeList rctl 
	    = sc.getCrossReferenceSourceInfo().getSubtypes(rct);
        ListOfKeYJavaType result = SLListOfKeYJavaType.EMPTY_LIST;
        for (int i=rctl.size()-1; i>=0 ; i--){
            rct = rctl.getClassType(i);
            KeYJavaType kct = (KeYJavaType)rec2key().toKeY(rct);
            result = result.prepend(kct);
        }
        return result;
    }

    private Recoder2KeY createRecoder2KeY() {
	return new Recoder2KeY(rec2key(), typeConverter);
    }

    /**
     * Parses a given JavaBlock using cd as context to determine the right
     * references.
     * @param block a String describing a java block
     * @param cd ClassDeclaration representing the context in which the
     * block has to be interpreted.
     * @return the parsed and resolved JavaBlock
     */

    public JavaBlock readBlock(String java, ClassDeclaration cd){
        return createRecoder2KeY().readBlock(java, new Context
            ((recoder.java.declaration.ClassDeclaration)
            rec2key().toRecoder(cd)));
    }


    /**
     * Parses a given JavaBlock using an empty context.
     * @param block a String describing a java block
     * @return the parsed and resolved JavaBlock
     */
    public JavaBlock readJavaBlock(String java) {
            return createRecoder2KeY().readBlockWithEmptyContext(java);
    }



    public void findImplementations(MethodCall mc,
                                    Type ct, String name,
                                    ListOfKeYJavaType signature) {

        recoder.service.CrossReferenceSourceInfo si 
	    = getServConf().getCrossReferenceSourceInfo();
        // set up recoder inputs
        recoder.abstraction.ClassType rct =
         (recoder.abstraction.ClassType) rec2key().toRecoder(ct);
        // transform the signature up to recoder conventions
        TypeArrayList rsignature = new TypeArrayList(signature.size());
        IteratorOfKeYJavaType i = signature.iterator();
        int j = 0;
        while (i.hasNext()) {
            rsignature.insert(j, (recoder.abstraction.Type)
            rec2key().toRecoder(i.next()));
            j++;
        }

        if (recFindImplementations(mc, rct, name, rsignature).contains(null)){
            // there was no implementation BELOW
            ClassTypeList superTypes = rct.getAllSupertypes();
            int k=0;
            while (!declaresApplicableMethods(superTypes.getClassType(k),
                                              name, rsignature)) k++;
            // we SHOULD have an implementation by now. Otherwise let it barf.
            rct = superTypes.getClassType(k);
            mc.addBranch(mc.makeIOf((KeYJavaType)mapping.toKeY(rct)), 
			 (KeYJavaType)mapping.toKeY(rct));

        }
    }



// method returns all classes that don't have an impl. below (yes?)
    private Vector recFindImplementations(MethodCall mc,
                                          recoder.abstraction.ClassType ct,
                                          String name,
                                          TypeList signature) {
        recoder.service.CrossReferenceSourceInfo si 
	    = getServConf().getCrossReferenceSourceInfo();
        Vector result = new Vector();
        ClassTypeList classes = si.getSubtypes(ct);
        boolean impBelow = false;


        // NULL in the vector means there is no impl. below
        for (int i = 0; i <classes.size(); i++) {
            recoder.abstraction.ClassType c = classes.getClassType(i);
            Vector v = recFindImplementations(mc, c, name, signature);
            impBelow = impBelow || !v.remove(null);
            result.addAll(v);
        }

        if (!impBelow) result.add(null);
        result.add(mapping.toKeY(ct));

        if (declaresApplicableMethods(ct, name, signature)) {
            if (impBelow)
                mc.addBranch(mc.disj(result), (KeYJavaType) mapping.toKeY(ct));
            else mc.addBranch(mc.makeIOf((KeYJavaType)mapping.toKeY(ct)),
			      (KeYJavaType)mapping.toKeY(ct));
                return new Vector();
        }

        return result;
    }



    private boolean declaresApplicableMethods(recoder.abstraction.ClassType ct,
                            String name,
                            TypeList signature) {
        recoder.service.CrossReferenceSourceInfo si 
	    = getServConf().getCrossReferenceSourceInfo();
	
	MethodList list = si.getMethods(ct);
    int s = list.size();
    int i = 0;
    while (i < s) {
        recoder.abstraction.Method m = list.getMethod(i);
        if (name.equals(m.getName())
        && si.isCompatibleSignature(signature, m.getSignature())
        && si.isVisibleFor(m, ct)) return true;
        else i++;
    }
    return false;
    }

    public KeYProgModelInfo copy() {
	return new KeYProgModelInfo(getServConf(), rec2key().copy(), 
				    typeConverter);
    }

}




