// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001 Universitaet Karlsruhe, Germany
//                    and Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General
// Public License. See LICENSE.TXT for details.
//

package de.uka.ilkd.key.java;
import de.uka.ilkd.key.java.declaration.*;
import de.uka.ilkd.key.java.declaration.modifier.*;
import de.uka.ilkd.key.java.expression.*;
import de.uka.ilkd.key.java.expression.operator.*;
import de.uka.ilkd.key.java.statement.*;
import de.uka.ilkd.key.java.reference.*;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.java.abstraction.Type;
import de.uka.ilkd.key.java.abstraction.PrimitiveType;
import de.uka.ilkd.key.java.abstraction.NullType;
import de.uka.ilkd.key.java.expression.literal.*;
import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.logic.*;
import java.lang.reflect.*;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Iterator;
import java.io.*;
import de.uka.ilkd.key.java.recoderext.KeYCrossReferenceSourceInfo;
import de.uka.ilkd.key.java.recoderext.KeYCrossReferenceServiceConfiguration;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.logic.op.*;
public class Recoder2KeY implements JavaReader{

    /**
     *   mapping from recoder.abstraction.Type to KeYJavaType
     */
    protected HashMap type2KeYType = new HashMap();


    /**
     *  methodsDeclaring contains the recoder method declarations as keys 
     *  that have been started to convert but are not yet finished.
     *  The mapped value is the reference to the later completed 
     *  ProgramMethod.
     */
    protected HashMap methodsDeclaring=new HashMap();

    protected KeYRecoderMapping rec2key = new KeYRecoderMapping();
    protected TypeConverter typeConverter;
    protected KeYCrossReferenceServiceConfiguration servConf;
    protected recoder.java.ProgramElement progElem;

    private Namespace knownSorts = new Namespace();

    private static int interactCounter=0;
   

    public Recoder2KeY(KeYRecoderMapping rec2key, TypeConverter tc) {
	this(new KeYCrossReferenceServiceConfiguration(),
	     null, null, rec2key, tc);
    }


    protected Recoder2KeY(KeYCrossReferenceServiceConfiguration servConf,
			  recoder.java.ProgramElement progElem,
			  String classPath,
			  KeYRecoderMapping rec2key,
			  TypeConverter tc) {
	this.servConf = servConf;
	this.progElem = progElem;
	this.typeConverter = tc;
	this.rec2key = rec2key;		
	if(classPath != null) {
	    servConf.getProjectSettings().setProperty
		(recoder.io.PropertyNames.INPUT_PATH,
		 classPath);
	}
	if (servConf.getProjectSettings().
	    ensureSystemClassesAreInPath()) {
	} else {
 	    Debug.out("System classes not found in path.");
	}

    }


    public Recoder2KeY(recoder.java.ProgramElement progElem,
		       String classPath, 
		       KeYRecoderMapping rec2key,
		       TypeConverter tc) {
	this(new KeYCrossReferenceServiceConfiguration(), 
	     progElem, classPath, rec2key, tc);
    }

    public Recoder2KeY(recoder.java.ProgramElement progElem,
		       KeYRecoderMapping rec2key,
		       TypeConverter tc) {
	this(progElem, null, rec2key, tc);
    }

    public Recoder2KeY(String classPath, KeYRecoderMapping rec2key,
		       TypeConverter tc) {
	this(null, classPath, rec2key, tc);
    }

     public Recoder2KeY(String classPath, Namespace knownSorts,
			KeYRecoderMapping rec2key,
			TypeConverter tc) {
	 this(null, classPath, rec2key, tc);
	 this.knownSorts = knownSorts;
     }

    private TypeConverter typeConverter() {
	return typeConverter;
    }

    /**
     * returns a javaInfo object with this mapping produced by recoder2key
     */
    public JavaInfo createJavaInfo(Services s) {
	return new JavaInfo(new KeYProgModelInfo(getServiceConfiguration(),
						 rec2key(),
						 typeConverter()), s);
    }


    public KeYCrossReferenceServiceConfiguration getServiceConfiguration() {
	return servConf;
    }


    public KeYRecoderMapping rec2key() {
	return rec2key;
    }

    
    /**
     * wraps a RECODER StatementBlock in a method
     * @param block the recoder.java.StatementBlock to wrap
     * @return the enclosing recoder.java.MethodDeclaration
     */
    protected recoder.java.declaration.MethodDeclaration 
	embedBlock(recoder.java.StatementBlock block) {

	/* 
	   MethodDeclaration(modifiers,return type,Identifier, parameters,
	                     throws, StatementBlock)
	 */
	recoder.java.declaration.MethodDeclaration mdecl=new 
	    recoder.java.declaration.MethodDeclaration
	    (null, null, new recoder.java.Identifier("SOMEMETHOD"), 
	     null, null, block);
	mdecl.makeParentRoleValid();
	return mdecl;
    }  


//     /**
//      * wraps a RECODER MethodDeclaration in a new class
//      * @param mdecl the recoder.java.declaration.MethodDeclaration to wrap
//      * @return the enclosing recoder.java.declaration.ClassDeclaration
//      */
//     protected recoder.java.declaration.ClassDeclaration
// 	embedMethod(recoder.java.declaration.MethodDeclaration mdecl
// 		    ) { 
	
// 	recoder.java.declaration.ClassDeclaration classDecl = 
// 	    new recoder.java.declaration.ClassDeclaration();

// 	return embedMethod(mdecl, classDecl, variableContext); 
//     }  

    static int count=0;

    /**
     * wraps a RECODER MethodDeclaration in a class
     * @param mdecl the recoder.java.declaration.MethodDeclaration to wrap
     * @param classContext the recoder.java.declaration.ClassDeclaration
     * where the method has to be embedded
     * @return the enclosing recoder.java.declaration.ClassDeclaration
     */
    protected recoder.java.declaration.ClassDeclaration
	embedMethod(recoder.java.declaration.MethodDeclaration mdecl,
		    Context context) {

	recoder.java.declaration.ClassDeclaration classContext = 
	    context.getClassContext();

	// add method to memberdeclaration list
	recoder.list.MemberDeclarationMutableList memberList = 
	    classContext.getMembers();
	if (memberList == null) {
	    memberList = new recoder.list.MemberDeclarationArrayList(1); 
	    classContext.setMembers(memberList);
	} 
	for (int i=0; i<memberList.size(); i++) {
	    if (memberList.getMemberDeclaration(i) 
		instanceof recoder.java.declaration.MethodDeclaration) {
		recoder.java.declaration.MethodDeclaration olddecl
		    = (recoder.java.declaration.MethodDeclaration) 
		    memberList.getMemberDeclaration(i);
		if (olddecl.getName().equals(mdecl.getName())) {
		    memberList.remove(i);
		}
	    }
	}
	memberList.add(mdecl);

	// add method to class

	//	classContext.setProgramModelInfo(servConf.getCrossReferenceSourceInfo());	
	classContext.makeAllParentRolesValid();
	
	return classContext;
    }  


//     protected recoder.java.CompilationUnit embedClass
// 	(recoder.java.declaration.ClassDeclaration classDecl) { 	
// 	return embedClass
// 	    (classDecl, 
// 	     new recoder.java.CompilationUnit());
//     }    


    /**
     * wraps a RECODER ClassDeclaration in a compilation unit
     * @param classDecl the recoder.java.ClassDeclaration to wrap
     * @param cUnit the recoder.java.CompilationUnit where the class is wrapped
     * @return the enclosing recoder.java.CompilationUnit
     */
    protected recoder.java.CompilationUnit embedClass
	(recoder.java.declaration.ClassDeclaration classDecl,
	 Context context) { 
	
	recoder.java.CompilationUnit cUnit = 
	    context.getCompilationUnitContext();

	// add class to compilation unit
	recoder.list.TypeDeclarationMutableList typeDecls = 	    
	    cUnit.getDeclarations();

	if (typeDecls==null) {
	    typeDecls=new recoder.list.TypeDeclarationArrayList(1); 
	} else {
	    typeDecls = (recoder.list.TypeDeclarationMutableList) 
		typeDecls.deepClone();
	}
	typeDecls.add(classDecl);

 	recoder.java.CompilationUnit compUnitContext
 	    = (recoder.java.CompilationUnit)cUnit.deepClone();
	
	compUnitContext.setDeclarations(typeDecls);
	compUnitContext.makeParentRoleValid();

	return compUnitContext;
    }  

    protected recoder.list.CompilationUnitMutableList recoderCompilationUnits
	(String[] cUnitStrings) {
	recoder.util.Debug.setLevel(500);
	recoder.list.CompilationUnitMutableList cUnits
	    = new recoder.list.CompilationUnitArrayList();	
	for (int i=0; i<cUnitStrings.length; i++) {
	    try {
		Debug.out("Reading "+cUnitStrings[i]);
		cUnits.add(servConf.getProgramFactory().parseCompilationUnit
		    (new StringReader(cUnitStrings[i]))); 
	    } catch (Exception e) {
		System.err.println("Trying to read "+cUnitStrings[i]
				   +" caused the "+
				   "exception:\n"+e);		
		//e.printStackTrace();
		throw new ConvertException(e.toString());
	    }
	}
	// run cross referencer
	for (int i=0; i<cUnits.size(); i++) {
	    servConf.getChangeHistory().attached(cUnits.getCompilationUnit(i));
	}
	servConf.getChangeHistory().updateModel();
	return cUnits;
    }

    protected recoder.list.CompilationUnitMutableList 
	recoderCompilationUnitsAsFiles(String[] cUnitStrings) {
	recoder.list.CompilationUnitMutableList cUnits
	    = new recoder.list.CompilationUnitArrayList();
	try {
	    cUnits=servConf.getProgramFactory().parseCompilationUnits(cUnitStrings);
	} catch (Exception e) {
		System.err.println("Trying to read "+cUnitStrings
				   +" caused the "+
				   "exception:\n"+e);		
		e.printStackTrace();
		throw new ConvertException(e.toString());
	}
	for (int i=0; i<cUnits.size(); i++) {
	    servConf.getChangeHistory().attached(cUnits.getCompilationUnit(i));
	}
	servConf.getChangeHistory().updateModel();
	return cUnits;
    }

    /**
     * adds a special compilation unit containing references to types that
     * have to be available in Recoder and KeY form, e.g. Exceptions
     */
    private void parseSpecial() {
	recoder.ProgramFactory pf=servConf.getProgramFactory();
	recoder.java.declaration.FieldDeclaration fd=null;
	try {
	    fd= pf.parseFieldDeclaration("Exception e;");
	} catch (Exception e) {
	    System.err.println("Error while parsing specials in R2K");
	}
	recoder.list.MemberDeclarationMutableList mds
	    = new recoder.list.MemberDeclarationArrayList();
	mds.add(fd);	
	recoder.java.Identifier id= new recoder.java.Identifier("KeYSpecialParsing");
	recoder.java.declaration.ClassDeclaration cd
	    = pf.createClassDeclaration(null, id, null, null, mds);
	recoder.list.TypeDeclarationMutableList tdl
	    = new recoder.list.TypeDeclarationArrayList(cd);
	recoder.java.CompilationUnit rcu=pf.createCompilationUnit(null, null, tdl);
	rcu.makeAllParentRolesValid();
	servConf.getChangeHistory().attached(rcu);
	servConf.getChangeHistory().updateModel();
	convert(rcu);
    }

    public CompilationUnit[] readCompilationUnits(String[] cUnitStrings) {
	recoder.list.CompilationUnitMutableList cUnits
	= recoderCompilationUnits(cUnitStrings);
	CompilationUnit[] result = new CompilationUnit[cUnitStrings.length];
	for (int i=0; i<cUnits.size(); i++) {
	    result[i]=(CompilationUnit) convert(cUnits.getCompilationUnit(i));
	}
	parseSpecial();
	return result;
    }

    public CompilationUnit[] readCompilationUnitsAsFiles(String[] cUnitStrings) {
	recoder.list.CompilationUnitMutableList cUnits
	= recoderCompilationUnitsAsFiles(cUnitStrings);
	CompilationUnit[] result = new CompilationUnit[cUnitStrings.length];
	for (int i=0; i<cUnits.size(); i++) {
	    Debug.out("R2K: ",cUnitStrings[i]);
	    result[i]=(CompilationUnit) convert(cUnits.getCompilationUnit(i));
	}
	parseSpecial();
	return result;
    }

    public CompilationUnit readCompilationUnit(String cUnitString) { 
	recoder.java.CompilationUnit cc=recoderCompilationUnits
	    (new String[]{cUnitString}).getCompilationUnit(0);
	parseSpecial();
	return (CompilationUnit) convert(cc);
    }


    private String trim(String s) {
	if (s.length()>100) return s.substring(0,150)+"[...]";
	return s;
    }


    /** parses a given JavaBlock using the context to determine the right 
     * references and returns a statement block of recoder.
     * @param block a String describing a java block
     * @param context recoder.java.CompilationUnit in which the block has 
     * to be interpreted
     * @return the parsed and resolved recoder statement block
     */
    protected recoder.java.StatementBlock 
	recoderBlock(String block,
		     Context context) { 
	recoder.java.StatementBlock bl = null;
	try {
	    bl = servConf.getProgramFactory().
		parseStatementBlock(new StringReader(block)); 
	} catch (ConvertException e) {
	    throw e;
	} catch (recoder.parser.ParseException pe) {
	    throw new ConvertException(pe);
	} catch (Exception e) {
	    System.err.println("Trying to read "+block+" caused the "+
			       "exception:\n"+e);	
	    if (block.length()>20) block=block.substring(0,20)+"...";
	    String message=trim(e.getMessage());
	    throw new ConvertException("When trying to read "+block
				       +": "+message);
	}
	context.getCompilationUnitContext().makeAllParentRolesValid();
	context.getClassContext().makeAllParentRolesValid();
	bl.makeAllParentRolesValid();
	embedMethod(embedBlock(bl), context);
	context.getCompilationUnitContext().makeAllParentRolesValid();
	servConf.getCrossReferenceSourceInfo().register(bl);
	try {
	    servConf.getChangeHistory().attached(context.getCompilationUnitContext());   	
	    servConf.getChangeHistory().updateModel();
	}  catch (ConvertException e) {
	    throw e;
	} catch (Exception e) {
	    if (block.length()>20) {
		block=block.substring(0,20)+"...";
	    }
	    throw new ConvertException("Probably "
				       +"an identifier has not been declared\n"
				       +"or no suitable Java model is available.");
	}
	return (recoder.java.StatementBlock)bl;
    } 

    /** parses a given JavaBlock using the context to determine the right 
     * references
     * @param block a String describing a java block
     * @param context recoder.java.CompilationUnit in which the block has 
     * to be interprested
     * @return the parsed and resolved JavaBlock
     */
    public JavaBlock readBlock(String block,
			       Context context) { 
	try {
	    try {
		JavaBlock jb = 
		    JavaBlock.createJavaBlock((JavaProgramElement)
					      callConvert(recoderBlock
							  (block, 
							   context)));	    
		return jb;
	    } catch (ConvertException e) {
		throw e;
	    } catch (NoSuchMethodException e) {
		JavaBlock jb2=null;
		try {jb2 = 
			 JavaBlock.createJavaBlock((JavaProgramElement)
					      convert(recoderBlock(block, 
								   context)));
		} catch (ConvertException e0) {
		    throw e0; 		
		} catch (Exception e0) {
		    throw new ConvertException(""+e0.getMessage());	    
		}
		return jb2;
	    }
	} catch (ConvertException e0) {
	    throw e0; 		
	}catch (Exception e) {
 	    throw new ConvertException(""+e);
	}
	//	return null;
    }

    /** parses a given JavaBlock using the context to determine the right 
     * references using an empty context
     * @param block a String describing a java block
     * @return the parsed and resolved JavaBlock
     */
    public JavaBlock readBlockWithEmptyContext(String block) { 

	return readBlock(block, createEmptyContext());

    }

    public JavaBlock readBlockWithProgramVariables(Namespace varns, String s) {
	IteratorOfNamed it=varns.allElements().iterator();
	ListOfProgramVariable pvs=SLListOfProgramVariable.EMPTY_LIST;
	while (it.hasNext()) {
	    Named n=it.next();
	    if (n instanceof ProgramVariable) {
		pvs=pvs.prepend((ProgramVariable)n);
	    }
	}
	Context ctx=createContext(pvs);
	return readBlock(s, ctx);
    }

    protected void insertToMap(recoder.java.ProgramElement r,
			     ProgramElement k) {
	if (r != null) {	    
	    rec2key.put(r, k);
	} else {
	    // commented out because they caused an excpetion (they were thrown at array references)
// 	    Debug.out("Rec2Key.insertToMap : Omitting entry  (r = " + r + " -> k = " + k + ")");
// 	    Debug.out("Maybe there is a bug !!");
	}
    }

    private void insertToMap(recoder.ModelElement r,
			     ModelElement k) {
	
	if (r != null) {
	    rec2key.put(r, k);
	} else {
	    // commented out because they caused an excpetion (they were thrown at array references)
// 	    Debug.out("Rec2Key.insertToMap : Omitting entry  (r = " + r + " -> k = " + k + ")");
// 	    Debug.out("Maybe there is a bug !!");
	}
    }


    /** determines the right convert method using reflection
     * @param pe the recoder.java.JavaProgramElement to be converted
     * @return the converted element
     */
    protected Object callConvert(recoder.java.JavaProgramElement pe)
	throws Exception{  
	return callConvert((recoder.java.ProgramElement) pe);
    }

    /** determines the right convert method using reflection
     * @param pe the recoder.java.JavaProgramElement to be converted
     * @return the converted element
     */
    protected Object callConvert(recoder.java.ProgramElement pe)
	throws Exception{ 
	Debug.assert(pe!=null);
	Method m=getClass().getDeclaredMethod("convert",
					      new Class[]{pe.getClass()});
	Object o=m.invoke(this, new Object[]{pe});
	
	return o;
    }

    /** calls the most specific convert method 
     * @param list an ExtList the converted element is added to
     * @param pe the recoder.java.JavaProgramElement to be converted
     */
    protected ExtList callConvertMethod(ExtList list,
					recoder.java.JavaProgramElement pe) {

	return callConvertMethod(list, (recoder.java.ProgramElement) pe);
    }
    /** calls the most specific convert method 
     * @param list an ExtList the converted element is added to
     * @param pe the recoder.java.JavaProgramElement to be converted
     */
    protected ExtList callConvertMethod(ExtList list,
					recoder.java.ProgramElement pe) {
	try {
	    list.add(callConvert(pe));
	} catch (ConvertException e) {
	    throw e;
	} catch (InvocationTargetException e) {
	    System.err.println("Exception while calling convert method for "+pe.toSource()+"/"+pe.getClass());
	    e.printStackTrace();
	    e.getTargetException().printStackTrace();
	    throw (ConvertException)(e.getTargetException());
	} catch (NoSuchMethodException e) {
	    try {
		list.add(convert(pe));
	    } catch (ConvertException e0) {
		throw e0; 		
	    } catch (Exception ex) {
		System.err.println("Converting "+pe+" failed due to a "+ex);
		throw new ConvertException(ex.getMessage());
	    }
	}  catch (Exception e) {
	    System.err.println("Converting "+pe+" failed due to a "+e);
	    throw new ConvertException(e.getMessage());
	} 
	return list;
    }

    /** calls the most specific convert method 
     * @param list an ExtList the converted element is added to
     * @param pe the recoder.java.JavaStatement to be converted
     */
    protected ExtList callConvertMethod(ExtList list,
					recoder.java.statement.JavaStatement pe) {
	return callConvertMethod(list, (recoder.java.Statement) pe);
    }

    /** calls the most specific convert method 
     * @param list an ExtList the converted element is added to
     * @param pe the recoder.java.Statement to be converted
     */
    protected ExtList callConvertMethod(ExtList list,
					recoder.java.Statement pe) {
	try {
	    list.add(callConvert(pe));
	} catch (ConvertException e) {
	    throw e;
	} catch (InvocationTargetException e) {
	    System.err.println("Exception while calling convert method for "+pe.toSource()+"/"+pe.getClass());
	    e.printStackTrace();
	    e.getTargetException().printStackTrace();
	    throw (ConvertException)(e.getTargetException());
	} catch (NoSuchMethodException e) {
	    try {
		list.add(convert(pe));
	    } catch (Exception ex) {
		System.err.println("Converting "+pe+" failed due to a "+ex);
	    }
	}  catch (Exception e) {
	    System.err.println("Converting "+pe+" failed due to a "+e);
	} 
	return list;
    }

    protected ProgramElement doConvert(recoder.java.JavaProgramElement pe) {
	ExtList el=new ExtList();
	callConvertMethod(el, pe);
	return (ProgramElement)el.get(0);
    }

    /** constructs the name of the corresponding KeYClass
     * @param recoderClass Class that is the original recoder 
     * @return String containing the KeY-Classname
     */
    protected String getKeYName(Class recoderClass) {
	return "de.uka.ilkd.key."
	    +recoderClass.getName().substring("recoder.".length());
    }

    
    /** gets the KeY-Class related to the recoder one
     * @param recoderClass the original Class
     * @return the related KeY Class 
     */
    protected Class getKeYClass(Class recoderClass) {
	try {
	    return Class.forName(getKeYName(recoderClass));
	} catch (Exception e) {
	    System.err.println("There is an AST class missing at KeY."+e);
	}
	return null;
    }

    /** determines the right standard constructor of the KeYClass
     * @param recoderClass the Class of the recoder AST object
     * @return the Constructor of the right KeY-Class
     */
    protected Constructor getKeYClassConstructor(Class recoderClass) {
	Constructor result=null;
	try {
	    result=getKeYClass(recoderClass).getConstructor(new
		Class[]{ExtList.class}); 
	} catch (Exception e) {
	    System.err.println("No matching constructor was found at "
			       +"KeY class "+getKeYClass(recoderClass)+".\n"
			       +e);
	    e.printStackTrace();
	}
	return result;
    }

//     /** collects children and adds their converted KeY-counterpart to
//      * the list of childrem
//      * @param pe the JavaNonTerminalProgramElement that needs its
//      * children before being converted
//      * @param children an ExtList with the children of pe
//      * @return the list of children after conversion
//      */
//     protected ExtList
// 	collectChildren(recoder.java.JavaNonTerminalProgramElement pe) { 
// 	return collectChildren((recoder.java.NonTerminalProgramElement) pe);
//     }

    /** collects children and adds their converted KeY-counterpart to
     * the list of childrem
     * @param pe the NonTerminalProgramElement that needs its
     * children before being converted
     * @param children an ExtList with the children of pe
     * @return the list of children after conversion
     */
    protected ExtList collectChildren
	(recoder.java.NonTerminalProgramElement pe) {
	ExtList children=new ExtList();
	for (int i=0; i<pe.getChildCount(); i++) {
	    callConvertMethod(children, 
			      (recoder.java.JavaProgramElement)
			      pe.getChildAt(i));

	}

	children.add(positionInfo(pe));
	return children;
    }

    protected PositionInfo positionInfo(recoder.java.SourceElement se) {
	Position relPos=new Position(se.getRelativePosition().getLine(),
				     se.getRelativePosition().getColumn());
	Position startPos=new Position(se.getStartPosition().getLine(),
				       se.getStartPosition().getColumn());
	Position endPos=new Position(se.getEndPosition().getLine(),
				     se.getEndPosition().getColumn());
	return new PositionInfo(relPos, startPos, endPos);
    }

    /** the standard case.
     * @param pe the recoder.java.JavaProgramElement to be converted
     * @return the converted de.uka.ilkd.key.java.JavaProgramElement
     */
    protected ProgramElement convert(recoder.java.JavaProgramElement pe) {
	return convert((recoder.java.ProgramElement) pe);
    }

    /** the standard case.
     * @param pe the recoder.java.ProgramElement to be converted
     * @return the converted de.uka.ilkd.key.java.JavaProgramElement
     */
    protected ProgramElement convert(recoder.java.ProgramElement pe) {
	
	ExtList children=new ExtList();
	if (pe instanceof recoder.java.NonTerminalProgramElement) {
	    children
		=collectChildren((recoder.java.JavaNonTerminalProgramElement)
				 pe);
	} else if (!(pe instanceof recoder.java.TerminalProgramElement)) {
	    throw new IllegalStateException
		(pe      
		 +"does not implement TerminalProgramElement "
		 +"or NonTerminalProgramElement. "
		 +" Change Recoder2Key by adding "
		 +"a new convert method for this class.");
			 
	}
	try {
	    //	    Debug.out("CHILDREN"+children+" CLASS"+pe.getClass());
	    ProgramElement result 
		= (ProgramElement) getKeYClassConstructor
		(pe.getClass()).newInstance(new Object[]{children});
	    return result;
	} catch (Exception e) {	    
	    System.err.println("Creation of a KeYASTNode failed. "
			       +"Tried to create a new "
			       +getKeYName(pe.getClass())
			       +"\nThe following exception had been "
			       +"thrown: "+e);
	    e.printStackTrace();
	    if (e instanceof java.lang.reflect.InvocationTargetException) {
		System.err.println("Target exception:"+
				   ((InvocationTargetException)e)
				   .getTargetException());
	    }
	    
	}
	return null;
    }
    
    /** converts the recoder.java.expression.operator.NewArray
     * node to the KeYDependance
     */
    protected NewArray convert(recoder.java.expression.operator.NewArray newArr){
	// first we need to collect all children
	ExtList children=collectChildren(newArr);
	// now we have to extract the array initializer
	// is stored separately and must not appear in the children list
	ArrayInitializer arrInit=(ArrayInitializer)
	    children.get(ArrayInitializer.class);
	children.remove(arrInit);	
	return new NewArray(children, arrInit, newArr.getDimensions());
    }


    //------------------- literals --------------------------------------
    

    /** convert a recoder IntLiteral to a KeY IntLiteral */
    protected IntLiteral convert(recoder.java.expression.literal.IntLiteral 
			       intLit) {
	// if there are comments to take into consideration 
	// change parameter to ExtList
	return new IntLiteral(intLit.getValue());
    }
    
    /** convert a recoder BooleanLiteral to a KeY BooleanLiteral */
    protected BooleanLiteral 
	convert(recoder.java.expression.literal.BooleanLiteral booleanLit) {

	// if there are comments to take into consideration 
	// change parameter to ExtList
	return (booleanLit.getValue() ? BooleanLiteral.TRUE : BooleanLiteral.FALSE);
    }


    /** convert a recoder StringLiteral to a KeY StringLiteral */
    protected StringLiteral 
	convert(recoder.java.expression.literal.StringLiteral stringLit) {

	// if there are comments to take into consideration 
	// change parameter to ExtList
	return new StringLiteral(stringLit.getValue());
    }

    /** convert a recoder DoubleLiteral to a KeY DoubleLiteral */
    protected DoubleLiteral 
	convert(recoder.java.expression.literal.DoubleLiteral doubleLit) {

	// if there are comments to take into consideration 
	// change parameter to ExtList
	return new DoubleLiteral(doubleLit.getValue());
    }

    /** convert a recoder FloatLiteral to a KeY FloatLiteral */
    protected FloatLiteral 
	convert(recoder.java.expression.literal.FloatLiteral floatLit) {

	// if there are comments to take into consideration 
	// change parameter to ExtList
	return new FloatLiteral(floatLit.getValue());
    }

    /** convert a recoder LongLiteral to a KeY LongLiteral */
    protected LongLiteral 
	convert(recoder.java.expression.literal.LongLiteral longLit) {

	// if there are comments to take into consideration 
	// change parameter to ExtList
	return new LongLiteral(longLit.getValue());
    }

    /** convert a recoder CharLiteral to a KeY CharLiteral */
    protected CharLiteral 
	convert(recoder.java.expression.literal.CharLiteral charLit) {

	// if there are comments to take into consideration 
	// change parameter to ExtList
	return new CharLiteral(charLit.getValue());
    }


    /** convert a recoder NullLiteral to a KeY NullLiteral */
    protected NullLiteral 
	convert(recoder.java.expression.literal.NullLiteral nullLit) {

	// if there are comments to take into consideration 
	// change parameter to ExtList
	return NullLiteral.NULL;
    }



    //----------------------------------------------------------

    
    /** convert a recoder EmptyStatement to a KeY EmptyStatement*/
    protected EmptyStatement convert(recoder.java.statement.EmptyStatement
				       eStmnt) {	
	// may change if comments are implemented, then 
	// new EmptyStatement(children);
	return new EmptyStatement();
    }
	
    /** convert a recoder EmptyStatement to a KeY EmptyStatement*/    
    protected ProgramElementName convert(recoder.java.Identifier id) {

	return new ProgramElementName(id.getText());
    }

    //------------------- operators -------------------------------- //

    protected Instanceof convert
	(recoder.java.expression.operator.Instanceof rio) {       
	return new Instanceof((Expression)doConvert
			      ((recoder.java.JavaProgramElement)
			       rio.getExpressionAt(0)),
			      (TypeReference)doConvert(rio.getTypeReference()));

    }

    //------------------- declaration -------------------------------- //
    
    protected InterfaceDeclaration convert
	(recoder.java.declaration.InterfaceDeclaration td) {

	KeYJavaType kjt =  getKeYJavaType(td);
	TypeDeclaration keyTD = new InterfaceDeclaration(collectChildren(td));	    
	kjt.setJavaType(keyTD);
	return (InterfaceDeclaration) kjt.getJavaType();	
    }
    
    


    protected ClassDeclaration convert
	(recoder.java.declaration.ClassDeclaration td) {
	KeYJavaType kjt = getKeYJavaType(td);
	ExtList classMembers = collectChildren(td);
	TypeDeclaration keyTD = new ClassDeclaration(classMembers);	    	    
	kjt.setJavaType(keyTD);
	return (ClassDeclaration) kjt.getJavaType();	
    }
            

    
    /** converts a recoder LocalVariableDeclaration to a KeY
     * LocalVariableDeclaration
     * (especially the declaration type of its parent is determined
     * and handed over)
     */
    protected LocalVariableDeclaration
	convert(recoder.java.declaration.LocalVariableDeclaration lvd) {  
	 LocalVariableDeclaration local = 
	     new LocalVariableDeclaration(collectChildren(lvd)); 

	 return local;
    }


    /** convert a recoder ParameterDeclaration to a KeY
     * ParameterDeclaration
     * (especially the declaration type of its parent is determined
     * and handed over)
     */
    protected ParameterDeclaration
	convert(recoder.java.declaration.ParameterDeclaration pd) {    
	 ParameterDeclaration param = 
	     new ParameterDeclaration(collectChildren(pd), pd.getASTParent()
				      instanceof 
				      recoder.java.declaration
				      .InterfaceDeclaration); 
	 return param;
    }


     /** convert a recoder FieldDeclaration to a KeY
      * FieldDeclaration
      * (especially the declaration type of its parent is determined
      * and handed over)
      */
     protected FieldDeclaration
 	convert(recoder.java.declaration.FieldDeclaration fd) {    
	 FieldDeclaration field = 
	     new FieldDeclaration(collectChildren(fd), fd.getASTParent()
				  instanceof recoder.java.declaration
				  .InterfaceDeclaration); 
	 return field;
     }

     /** convert a recoder ConstructorDeclaration to a KeY
      * ConstructorDeclaration
      * (especially the declaration type of its parent is determined
      * and handed over)
      */
     protected ConstructorDeclaration
 	convert(recoder.java.declaration.ConstructorDeclaration cd) {    

	 ConstructorDeclaration consDecl
	     =new ConstructorDeclaration
	     (collectChildren(cd), cd.getASTParent() instanceof
	      recoder.java.declaration.InterfaceDeclaration);
	insertToMap(cd, consDecl);
	return consDecl;
     }


    protected TypeCast
	convert(recoder.java.expression.operator.TypeCast c) {
	ExtList children=collectChildren(c);
	TypeReference tr=(TypeReference)children.get(0);
	Expression e=(Expression)children.get(1);
	return new TypeCast(e, tr);
	
    }

    
    private KeYJavaType lookup(recoder.abstraction.Type t) {
	return (KeYJavaType) rec2key.toKeY(t);
    }
    

    private LinkedList directSuperClasses
	(recoder.java.declaration.TypeDeclaration tdl){

	recoder.service.CrossReferenceSourceInfo crsc
	    = getServiceConfiguration().getCrossReferenceSourceInfo();
	LinkedList sc=new LinkedList();
	if (tdl instanceof recoder.java.declaration.InterfaceDeclaration) {
	    recoder.java.declaration.Extends exts 
		= ((recoder.java.declaration.InterfaceDeclaration)
		   tdl).getExtendedTypes();
	    for (int i=0; i<exts.getTypeReferenceCount(); i++) {
		recoder.java.reference.TypeReference ref 
		    = exts.getTypeReferenceAt(i);
		sc.add((recoder.abstraction.ClassType) crsc.getType(ref));
	    }
	    return sc;
	} else if (tdl instanceof recoder.java.declaration.ClassDeclaration) {
	    recoder.java.declaration.Implements impls 
		= ((recoder.java.declaration.ClassDeclaration)tdl)
		.getImplementedTypes();
	    for (int i=0; i<impls.getTypeReferenceCount(); i++) {
		recoder.java.reference.TypeReference ref 
		    = impls.getTypeReferenceAt(i);
		sc.add((recoder.abstraction.ClassType) crsc.getType(ref));
	    }
	    sc.add((recoder.abstraction.ClassType) crsc.getType
		      (((recoder.java.declaration.ClassDeclaration)tdl)
		       .getExtendedTypes()
		       .getTypeReferenceAt(0)));
	    return sc;
	}
	Debug.assert(false);
	return null;
    }

    private boolean isObject(recoder.abstraction.ClassType ct) {
	return "java.lang.Object".equals(ct.getFullName())
	    || "Object".equals(ct.getName());
    }

    private void createObjectSort(recoder.abstraction.ClassType ct, 
				  SetOfSort supers) {
	Sort newSort=null;
	if (!isObject(ct)) {
	    newSort = new ObjectSort(new Name(ct.getFullName()), supers);
	} else {
	    newSort = typeConverter().getObjectSort();
	}
	addKeYJavaType(ct, newSort);
    }


    private boolean containsType(recoder.abstraction.Type t) {
	return rec2key.mapped(t);
    }

    private SetOfSort directSuperSorts
	(recoder.abstraction.ClassType classType) {

	recoder.list.ClassTypeList supers=classType.getSupertypes();
	SetOfSort ss=SetAsListOfSort.EMPTY_SET;
	for (int i=0; i<supers.size(); i++) {
	    if (!containsType(supers.getClassType(i))) {
		createObjectSort(supers.getClassType(i), 
				 directSuperSorts(supers.getClassType(i)));
	    }
	    ss=ss.add((Sort)getKeYJavaType(supers.getClassType(i)).getSort());	    
	}
	if (ss==SetAsListOfSort.EMPTY_SET && !isObject(classType)) {
	    ss=ss.add(typeConverter().getObjectSort());
	}
	return ss;
    }

    private KeYJavaType getKeYJavaType(recoder.abstraction.Type t) {		
	if (t==null) {
	    return null; //this can be originate from 'void'
	}
	KeYJavaType kjt = lookup(t);	    
	if (kjt != null) {
	    return kjt;
	}
	// create a new KeYJavaType
	if (t instanceof recoder.abstraction.PrimitiveType) {
	    Sort s = typeConverter.getPrimitiveSort
		(PrimitiveType.getPrimitiveType(t.getFullName()));
	    if (s==null) {
		System.err.println("Warning: Type "+t.getFullName()+" not supported");
		s=new PrimitiveSort(new Name(t.getFullName()));
	    }
	    addKeYJavaType(t, s);
	} else if (t instanceof recoder.abstraction.ClassType) {
	    recoder.abstraction.ClassType ct=(recoder.abstraction.ClassType)t;
	    createObjectSort(ct, directSuperSorts(ct));	    
	} else if (t instanceof recoder.abstraction.ArrayType){
	    System.err.println("Warning: ArrayType not supported");
	    Sort s=new PrimitiveSort(new Name(t.getFullName()));
	    addKeYJavaType(t, s);
	}
	return lookup(t);
    }
    
    private KeYJavaType addKeYJavaType(recoder.abstraction.Type t, Sort s) {
	KeYJavaType result = null;       
	if (!(t instanceof recoder.java.declaration.TypeDeclaration)) {
	    Type type = null;
	    if (t instanceof recoder.abstraction.PrimitiveType) {
		type = PrimitiveType.getPrimitiveType(t.getFullName());
		result = typeConverter.getKeYJavaType(type);		
		if (result == null) {
		    Debug.out("recoder2key: create new KeYJavaType for", t);
		    Debug.out("recoder2key: this should not happen");
		    result = new KeYJavaType(type, s);
		}
	    } else if (t instanceof recoder.abstraction.NullType) {
		type = NullType.JAVA_NULL;
		result = new KeYJavaType(type);
	    } else if (t instanceof recoder.bytecode.ClassFile) {
		    recoder.bytecode.ClassFile cf = (recoder.bytecode.ClassFile)t;
		    if (!cf.getName().equals("Object")) {
			type = new ClassDeclaration(cf.getName());
			result = new KeYJavaType(type, s);
		    } else {			
			result = typeConverter().getObjectKeYJavaType();
		    }
	    } else {
		Debug.out("recoder2key: unknown type", t);
		result = new KeYJavaType();
	    }
	    	    	    
	} else {
	    knownSorts.add(s);
	    if (s!=typeConverter().getObjectSort()) {
		result = new KeYJavaType(s);		    
	    } else {
		result = typeConverter().getObjectKeYJavaType();
	    }
	}
	insertToMap(t, result);
	return result;
    }

    // ----------------- references ----------------------------------- //


    /** convert a recoder VariableSpecification to a KeY
     * VariableSpecification
     * (checks dimension and hands it over and insert in hashmap)
     */
    protected VariableSpecification
	convert(recoder.java.declaration.VariableSpecification recoderVarspec){	

	VariableSpecification varSpec
	    = (VariableSpecification)rec2key.toKeY(recoderVarspec);
	if (varSpec == null) {
	    recoder.abstraction.Type recoderType = 
		(servConf.getSourceInfo()).getType(recoderVarspec);

	    KeYJavaType kjt = getKeYJavaType(recoderType);
	    ProgramElementName name = new ProgramElementName
		(recoderVarspec.getName());
	    ProgramVariable progvar = new ProgramVariable(name, kjt);
	    ExtList children = collectChildren(recoderVarspec);
	    varSpec = new VariableSpecification(children, 
						progvar,
						recoderVarspec.getDimensions(),
						kjt);
	    insertToMap(recoderVarspec, varSpec);
	}
	return varSpec;
    }


    /** convert a recoder MethodDeclaration to a KeY
     * ProgramMethod
     * (especially the declaration type of its parent is determined
     * and handed over)
     */
    protected ProgramMethod
	convert(recoder.java.declaration.MethodDeclaration md) {     

	ProgramMethod result=null;

	// methodsDeclaring contains the recoder method declarations as keys 
	// that have been started to convert but are not yet finished.
	// The mapped value is the reference to the later completed 
	// ProgramMethod.
	if (methodsDeclaring.containsKey(md)) {
	    // a recursive call from a method reference
	    return (ProgramMethod) methodsDeclaring.get(md);
	                                    //reference that will later be set.
	} 
	methodsDeclaring.put(md, result);
	if (!rec2key.mapped(md)) {
	    MethodDeclaration methDecl
		=new MethodDeclaration(collectChildren(md), md.getASTParent()
				       instanceof
				       recoder.java.declaration
				       .InterfaceDeclaration);
	    KeYJavaType kjt  = getKeYJavaType(md.getReturnType());
	    recoder.abstraction.ClassType cont
		= servConf.getCrossReferenceSourceInfo()
	            .getContainingClassType((recoder.abstraction.Member)md);
	    KeYJavaType contKJT=getKeYJavaType(cont);
	    result = new ProgramMethod(methDecl, contKJT, kjt);
	    insertToMap(md, result);
	}
	methodsDeclaring.remove(md);
	result=(ProgramMethod)rec2key.toKeY(md);
	return result;
    }


    /** convert a recoder FieldSpecification to a KeY
     * FieldSpecification
     * (checks dimension and hands it over and insert in hash map)
     */
    protected FieldSpecification
 	convert(recoder.java.declaration.FieldSpecification recoderVarspec){

	if (recoderVarspec==null) { //%%%%%%%%%%%%%
	    return new FieldSpecification();
	}

	FieldSpecification varSpec
	    = (FieldSpecification)rec2key.toKeY(recoderVarspec);
	KeYJavaType kjt;
	ProgramElementName name; 
	int dim=0;
	if (varSpec==null) {
	    recoder.abstraction.Type recoderType = 
		(servConf.getSourceInfo()).getType(recoderVarspec);
	    
	    kjt  = getKeYJavaType(recoderType);
	    name = new ProgramElementName(recoderVarspec.getName()); 
	    dim  = recoderVarspec.getDimensions();
	    ProgramVariable progvar = new ProgramVariable(name, kjt);
	    varSpec = new FieldSpecification(collectChildren(recoderVarspec), 
					     progvar, dim, kjt);
	    insertToMap(recoderVarspec, varSpec);
	}
	return varSpec;
    }


    /** convert a recoder TypeReference to a KeY
     * TypeReference
     * (checks dimension and hands it over)
     */
    protected TypeReference
	convert(recoder.java.reference.TypeReference tr) {
	recoder.abstraction.Type rType
	    = getServiceConfiguration().getSourceInfo().getType(tr);
	if (rType==null) return null; // because of 'void'
	KeYJavaType kjt=getKeYJavaType(rType);	
	TypeReference result = new TypeRef(collectChildren(tr), 
					   tr.getDimensions(), kjt); 
	return result;
    }

    /** if an UncollatedReferenceQualifier appears throw a
     * ConvertExceception because these qualifiers have to be resolved
     * by running the CrossReferencer
     */
    protected JavaProgramElement 
	convert(recoder.java.reference.UncollatedReferenceQualifier urq) {    
	throw new ConvertException("Unknown qualifiers appeared. ");
    }


    protected recoder.java.declaration.VariableSpecification 
	getRecoderVarSpec(recoder.java.reference.VariableReference vr) {
	recoder.service.SourceInfo 
	    sourceInfo = servConf.getSourceInfo();
	return sourceInfo.getVariableSpecification(sourceInfo.getVariable(vr));
    }

    /**
     * converts a recoder variable reference. A ProgramVariable is created
     * replacing the variable reference.
     * @param fr the recoder variable reference.
     */
     protected ProgramVariable convert
	 (recoder.java.reference.VariableReference vr) {

	 recoder.java.declaration.VariableSpecification recoderVarspec
	     = getRecoderVarSpec(vr);	 
	 if (rec2key.mapped(recoderVarspec)) {
	     VariableSpecification vs
		 = (VariableSpecification) rec2key.toKeY(recoderVarspec);
	     ProgramVariable pv = (ProgramVariable)vs.getProgramVariable();
	     return pv;
	 } else {
	     VariableSpecification keyVarspec = convert(recoderVarspec);
	     ProgramVariable progvar = (ProgramVariable) 
		 keyVarspec.getProgramVariable();
	     insertToMap(recoderVarspec, keyVarspec);
	     return progvar;	     
	 }
     }


    /**
     * converts a recoder field reference. A ProgramVariable is created
     * replacing the field reference.
     * @param fr the recoder field reference.
     */
     protected FieldReference convert
	 (recoder.java.reference.FieldReference fr) {
	 ProgramVariable pv;
	 recoder.java.declaration.FieldSpecification recoderVarspec
	     = (recoder.java.declaration.FieldSpecification)
	         getRecoderVarSpec(fr);

	 ReferencePrefix prefix=null;	
	 if (fr.getReferencePrefix() != null) {
	     prefix = (ReferencePrefix)doConvert
		 ((recoder.java.JavaProgramElement)fr.getReferencePrefix());
	 }


	 if (recoderVarspec == null) { 
	     // null means only bytecode available for this 
	     // field %%%
 	    recoder.abstraction.Type recoderType = 
 		servConf.getByteCodeInfo().getType
		(servConf.getSourceInfo().getField(fr));	    

	     recoderVarspec = new recoder.java.declaration.
		 FieldSpecification(fr.getIdentifier());

	    KeYJavaType kjt =  getKeYJavaType(recoderType);

	     pv = new ProgramVariable
		 (new ProgramElementName(fr.getName()), kjt);
	     FieldSpecification keyVarspec = 
		 new FieldSpecification(pv);     	     
	     insertToMap(recoderVarspec, keyVarspec);	     
	     return new FieldReference(pv, prefix);
	 } 

	 if (rec2key.mapped(recoderVarspec)) {
	     VariableSpecification vs
		 = (VariableSpecification) rec2key.toKeY(recoderVarspec);
	     pv = (ProgramVariable) vs.getProgramVariable();	     
	 } else {
	     VariableSpecification keyVarspec = convert(recoderVarspec);
	     pv = (ProgramVariable) keyVarspec.getProgramVariable();
	     insertToMap(recoderVarspec, keyVarspec);
	 }


	 recoder.abstraction.ClassType ct = 
	     servConf.getSourceInfo().getContainingClassType(fr);

	 if (ct != null && 
 	     ct.getFullName().startsWith("KeY_interactive_input_")) {
 	     return new EmbededVariableReference(pv, prefix);
 	 } else {
	     return new FieldReference(pv, prefix);
 	 }
     }



    /**
     * converts a recoder method reference. A
     * de.uka.ilkd.key.logic.op.ProgramMethod is created 
     * replacing the method reference.
     * @param mr the recoder method reference.
     * @return the Method the KeY Dependance
     */
     protected MethodReference convert
	 (recoder.java.reference.MethodReference mr) {

	 recoder.service.SourceInfo 
	     sourceInfo = servConf.getSourceInfo();
	 recoder.abstraction.Method method = null;
	 try {
	     method=sourceInfo.getMethod(mr);
	 } catch (Exception e) {
	     System.err.println("An error occured while trying to "+
				"resolve a method reference. Probably"+
				" caused by declaring two methods with"+
				"same signature.");
	     e.printStackTrace();
	     throw new ConvertException(e.toString());
	 }


	 if (!rec2key.mapped(method)) 
	     if (method instanceof recoder.java.declaration.MethodDeclaration) {
		 // method reference before method decl, also recursive calls.
		 //	do not use: 
		 ProgramMethod keyMethodDecl
		    =convert((recoder.java.declaration.MethodDeclaration)method);
		 // because of cycles when reading recursive programs

	 }


	 ExtList children=collectChildren(mr);
	 // convert reference prefix separately
	 ReferencePrefix prefix=null;
	 int prefixPos=getPosition(mr, mr.getReferencePrefix());
	 if (prefixPos!=-1) {
	     prefix=(ReferencePrefix)children.get(prefixPos);
	     children.remove(prefixPos);
	 }
	 return new
	     de.uka.ilkd.key.java.reference.MethodReference
	     (children,prefix,(ProgramMethod)rec2key.toKeY(method));

     }



    //--------------Special traetment because of ambiguities ----------


    protected LabeledStatement 
	convert(recoder.java.statement.LabeledStatement l) {

	ExtList children=collectChildren(l);
	Label lab=null;
	int labPos=getPosition(l,l.getIdentifier());	
	if (labPos!=-1) {
	    lab=(Label)children.get(labPos);
	    children.remove(labPos);
	}
	return new LabeledStatement(children, lab);
    }

    /**
     * converts a For.
     * @param f the For of recoder
     * @return the For of KeY
     */
    protected For convert(recoder.java.statement.For f) {
	return new For(convertLoopInitializers(f), 
		       convertGuard(f), 
		       convertUpdates(f), 
		       convertBody(f));
    }

    /**
     * converts a While.
     * @param f the While of recoder
     * @return the While of KeY
     */
    protected While convert(recoder.java.statement.While w) {
	return new While(convertGuard(w).getExpression(), 
			 convertBody(w));
    }

    /**
     * converts a Do.
     * @param f the Do of recoder
     * @return the Do of KeY
     */
    protected Do convert(recoder.java.statement.Do d) {
	return new Do(convertGuard(d).getExpression(), 
		      convertBody(d));
    }

    /**
     * helper for convert(x) with x a LoopStatement. Converts the body of x.
     */
    protected Statement convertBody(recoder.java.statement.LoopStatement ls) {
	Object body=null;
	if (ls.getBody()!=null) {
	    body=callConvertMethod(new ExtList(), 
				   (recoder.java.JavaProgramElement)
				   ls.getBody()).get(0);
	}
	return (Statement)body;
    }

    /**
     * helper for convert(x) with x a LoopStatement. Converts the guard of x.
     */
    protected Guard convertGuard(recoder.java.statement.LoopStatement ls) {
	Object guard=null;
	if (ls.getGuard()!=null) {
	    guard=callConvertMethod(new ExtList(), 
				    (recoder.java.JavaProgramElement)
				    ls.getGuard()).get(0);
	}
	return new Guard((Expression)guard);
    }

    /**
     * helper for convert(x) with x a LoopStatement. Converts the updates of x.
     */
    protected ForUpdates
	convertUpdates(recoder.java.statement.LoopStatement ls) {
	    ExtList updates=new ExtList();
	    if (ls.getUpdates()!=null) {
		for (int i=0; i<ls.getUpdates().size(); i++) {
		    updates=callConvertMethod(updates, 
					      (recoder.java.JavaProgramElement)
					     ls.getUpdates().getExpression(i));
		}	    
	    } else {
		return null;
	    }
	    return new ForUpdates(updates);
	}

    /**
     * helper for convert(x) with x a LoopStatement. Converts the loop
     * initializers of x. 
     */
    protected LoopInit
	convertLoopInitializers(recoder.java.statement.LoopStatement ls) {
	    ExtList inits=new ExtList();
	    if (ls.getInitializers()!=null) {
		for (int i=0; i<ls.getInitializers().size(); i++) {
		    inits
			=callConvertMethod(inits, 
					   (recoder.java.JavaProgramElement)
					   (ls.getInitializers()
					   .getLoopInitializer(i)));

		}	    
	    } else {
		return null;	 
	    }
	   
	    LoopInitializer[] result=new LoopInitializer[inits.size()];
	    for (int j=0; j<inits.size(); j++) {
		result[j]=(LoopInitializer)inits.get(j);
	    }
	    return new LoopInit(result);
	}

 

    /**
     * converts an ArrayReference. Special handling because the initializing
     * Expressions and the ReferencePrefix accessPath might not be disjunct.
     */
    protected ArrayReference convert(recoder.java.reference.ArrayReference ar) {
	ExtList children=collectChildren(ar);
	ReferencePrefix prefix=null;

	int prefixPos=getPosition(ar,ar.getReferencePrefix());
	if (prefixPos!=-1) {
	    prefix=(ReferencePrefix)children.get(prefixPos);
	    children.remove(prefixPos);
	}
	return new ArrayReference(children, prefix);
    }




    /**
     * converts a SpecialConstructorReference. 
     * Special handling because the initializing
     * Expressions and the ReferencePrefix accessPath might not be disjunct.
     */
    protected SuperConstructorReference 
	convert(recoder.java.reference.SuperConstructorReference scr) {
	ExtList children=collectChildren(scr);
	ReferencePrefix prefix=null;
	int prefixPos=getPosition(scr,scr.getReferencePrefix());	     
	if (prefixPos!=-1) {
	    prefix=(ReferencePrefix)children.get(prefixPos);
	    children.remove(prefixPos);
	}
	return new SuperConstructorReference(children, prefix);
    }


    /**
     * converts a Case. 
     * Special handling because the initializing
     * Expression and Statements might not be disjunct.
     */
    protected Case 
	convert(recoder.java.statement.Case c) {
	ExtList children=collectChildren(c);
	Expression expr=null;
	int exprPos=getPosition(c,c.getExpression());	
	if (exprPos!=-1) {
	    expr=(Expression)children.get(exprPos);
	    children.remove(exprPos);
	}
	return new Case(children, expr);
    }


    /**
     * converts a New. 
     * Special handling because the ReferencePrefix and the TypeReference
     *  might not be disjunct.
     */
    protected New 
	convert(recoder.java.expression.operator.New n) {

	ExtList children=collectChildren(n);
	ReferencePrefix rp=null;
	int rpPos=getPosition(n,n.getReferencePrefix());
	if (rpPos!=-1) {
	    rp=(ReferencePrefix)children.get(rpPos);
	    children.remove(rpPos);
	}
	return new New(children, rp);
    }


    //-----------------------------------------------------------


    protected Import convert(recoder.java.Import im) {
	return new Import(collectChildren(im), 
			  im.isMultiImport());
    }

    /**
     * returns the position of pe2 in the virtual child array of pe1
     */
    protected static int getPosition(recoder.java.NonTerminalProgramElement 
				   pe1, 
				   recoder.java.ProgramElement pe2) {
	int n=pe1.getChildCount();
	int i=0;
	while ((i<n) && (pe1.getChildAt(i)!=pe2)) i++;
	return (i==n) ? -1 : i; 
    }

    
    public JavaProgramElement convert() {
	return (JavaProgramElement)
	    callConvertMethod(new ExtList(),progElem).getFirst();
    }

    private recoder.java.declaration.ClassDeclaration interactClassDecl() {
	recoder.java.declaration.ClassDeclaration classContext = 
 	    new recoder.java.declaration.ClassDeclaration
  		(null, new recoder.java.Identifier("KeY_interactive_input_"
						   +interactCounter++), 
		 null, null, null);
 	classContext.setProgramModelInfo(servConf.getCrossReferenceSourceInfo());
	return classContext;
    }

    /** creates an empty RECODER compilation unit 
     * @return the recoder.java.CompilationUnit 
     */
    public Context createEmptyContext() {
 	recoder.java.declaration.ClassDeclaration classContext 
	    = interactClassDecl();
 	return new Context(classContext);

	//		return new Context();
    }

    private VariableSpecification lookupVarSpec(ProgramVariable pv) {
	Iterator it=rec2key.elemsKeY().iterator();
	while (it.hasNext()) {
	    Object o=it.next();
	    if ((o instanceof VariableSpecification) 
		&& ((VariableSpecification)o).getProgramVariable()==pv){
		
		return (VariableSpecification)o;
	    }
	}
	return null;
    }


    public void addProgramVariablesToClassContext
	(recoder.java.declaration.ClassDeclaration classContext,
	 ListOfProgramVariable vars, 
	 recoder.service.CrossReferenceSourceInfo csi) {
	
	IteratorOfProgramVariable it = vars.iterator();
	java.util.HashSet names=new java.util.HashSet();
	recoder.list.MemberDeclarationMutableList list = classContext.getMembers();
	if (list == null) {
	    list = new recoder.list.MemberDeclarationArrayList();
	    classContext.setMembers(list);
	}
	l: while (it.hasNext()) {
 	    VariableSpecification keyVarSpec;
	    ProgramVariable var = it.next();
	    if (names.contains(var.name().toString())) {
		continue l;}
	    names.add(var.name().toString());
	    keyVarSpec=lookupVarSpec(var);
	    if (keyVarSpec==null) {
		keyVarSpec=new FieldSpecification(var);
	    }
	    recoder.java.declaration.FieldDeclaration recVar=
		new recoder.java.declaration.FieldDeclaration
		    (null, new recoder.java.reference.TypeReference
			(new recoder.java.Identifier
			    (var.getKeYJavaType().getFullName())),
		     new recoder.java.Identifier(keyVarSpec.getName()),
		     null);
 	    list.add(recVar);
 	    classContext.makeAllParentRolesValid();
	    recoder.java.declaration.VariableSpecification rvarspec
		= recVar.getVariables().getVariableSpecification(0);
	    rvarspec.setProgramModelInfo(csi);
	    insertToMap(recVar.getVariables()
			.getVariableSpecification(0), keyVarSpec);
	    //	    insertToMap(recVar, keyVarspec);
	}
	servConf.getChangeHistory().updateModel();
    }

    

    public Context createContext(ListOfProgramVariable pvs) {
	return createContext(pvs, servConf.getCrossReferenceSourceInfo());
    }

    public Context createContext(ListOfProgramVariable vars, 
				 recoder.service.CrossReferenceSourceInfo csi) {
 	recoder.java.declaration.ClassDeclaration classContext 
	    = interactClassDecl();
	addProgramVariablesToClassContext(classContext, vars, csi);	
	return new Context(classContext);
    }


    

    
    class KeYIdentifierSnd extends recoder.java.Identifier {
	KeYIdentifierSnd(String id) {
	    this.id = id;
	}
    }

}
