// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001 Universitaet Karlsruhe, Germany
//                    and Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General
// Public License. See LICENSE.TXT for details.
//

package de.uka.ilkd.key.java;




/**
 * this is a collection of common services to the KeY prover. Services
 * include information on the underlying Java model and a converter to
 * transform Java program elements to logic (where possible) and back.
 */
public class Services{


    /** used to convert types, expressions and so on to logic elements
     * (in special into to terms or formulas)
     */
    private TypeConverter typeconverter;

    /**
     * the information object on the Java model
     */
    private JavaInfo javainfo;
    
    /**
     * creates a new Services object with a new TypeConverter and a new
     * JavaInfo object with no information stored at none of these.
     */
				  
    public Services(){
        typeconverter = new TypeConverter(this);
	KeYProgModelInfo kpmi 
	    = new KeYProgModelInfo(new KeYRecoderMapping(), 
				   typeconverter);
        javainfo = new JavaInfo(kpmi, this);
    }

    /**
     * creates a Services object that contains the given JavaInfo object, a
     * type converter is created
     * @param ji the JavaInfo to use
     */
    public Services(JavaInfo ji){
        javainfo = ji;
        typeconverter = new TypeConverter(this);
    }

    /**
     * Returns the TypeConverter associated with this Services object.
     */
    public TypeConverter getTypeConverter(){
        return typeconverter;
    }


    public void setTypeConverter(TypeConverter tc){
        typeconverter=tc;
    }


    /**
     * Returns the JavaInfo associated with this Services object.
     */
    public JavaInfo getJavaInfo(){
        return javainfo;
    }


    public void setJavaInfo(JavaInfo ji) {
	javainfo=ji;
    }

    /**
     * creates a new services object containing a copy of the java info of
     * this object and a new TypeConverter (deep copy)
     * @return the copy
     */
    public Services copy() {
	Services s = new Services();
	s.setJavaInfo(getJavaInfo().copy(s));
	return s;
    }
	
}
	
	
	
	

