// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.java.abstraction;

import java.util.HashMap;

/**
   A program model element representing primitive types.
   @author AL
   @author RN
 */
public class PrimitiveType
    implements Type {

    public static final PrimitiveType JAVA_BYTE = new PrimitiveType("byte");
    public static final PrimitiveType JAVA_SHORT = new PrimitiveType("short");
    public static final PrimitiveType JAVA_INT = new PrimitiveType("int");
    public static final PrimitiveType JAVA_LONG = new PrimitiveType("long");
    public static final PrimitiveType JAVA_FLOAT = new PrimitiveType("float");
    public static final PrimitiveType JAVA_DOUBLE = new PrimitiveType("double");
    public static final PrimitiveType JAVA_BOOLEAN = new PrimitiveType("boolean");
    public static final PrimitiveType PROGRAM_SV = new PrimitiveType("SV");
    private static final HashMap typeMap = new HashMap(); 
    static {
	typeMap.put("byte", JAVA_BYTE);
	typeMap.put("short", JAVA_SHORT);
	typeMap.put("int", JAVA_INT);
	typeMap.put("long", JAVA_LONG);
	typeMap.put("float", JAVA_FLOAT);
	typeMap.put("double", JAVA_DOUBLE);
	typeMap.put("boolean", JAVA_BOOLEAN);	
    }

    public static PrimitiveType getPrimitiveType(String name) {
	return (PrimitiveType)typeMap.get(name);
    }

    private String name;

    private PrimitiveType(String name) {
	this.name = name.intern();
    }

    /**
       Returns the name of this type.
       @return the name of this type.
     */
    public String getName() {
	return name;
    }

    public boolean equals(Object o) {
	if (o instanceof PrimitiveType &&
	    ((PrimitiveType)o).getName().equals(name)) {
		return true;
	}
	return false;
    }
    
    /**
       Returns the name of type.
       @return the full name of this program model element.
     */
    public String getFullName() {
	return name;
    }

    /**
       Returns the name of type.
       @return the full name of this program model element.
     */
    public String toString() {
	return "PrimitiveType:"+name;
    }

}
