// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//


package de.uka.ilkd.key.java.reference;

import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.declaration.*;
import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.java.visitor.Visitor;

/**
 *  Super constructor reference.
 *  
 */

public class SuperConstructorReference
    extends SpecialConstructorReference
    implements ReferenceSuffix {

   
    /**
     Access path.
     */
    protected final ReferencePrefix prefix;


    /**
     *      Super constructor reference.
     */
    public SuperConstructorReference() {
	prefix =null;
    }

    /**
     *      Super constructor reference.
     *      @param arguments an expression mutable list.
     */
    public SuperConstructorReference(Expression[] arguments) {
        super(arguments);
	prefix =null;
    }

    /**
     *      Super constructor reference.
     *      @param path a reference prefix.
     *      @param arguments an expression mutable list.
     */

    public SuperConstructorReference(ReferencePrefix path,
                                     Expression[] arguments) {
        super(arguments);
        prefix =path;
    }

    /**
     * Constructor for the transformation of COMPOST ASTs to KeY.
     * @param children the children of this AST element as KeY classes.
     *  May contain: 
     * 	several of Expression (as initializers of the array),
     * 	Comments.
     *  MUST NOT CONTAIN: the ReferencePrefix for the accessPath because
     *    Expression and ReferencePrefix might not be disjunct.
     * @param accessPath a ReferencePrefix of the array reference. 

     */ 
    public SuperConstructorReference(ExtList children, 
				     ReferencePrefix accessPath) {
	super(children);
	this.prefix=accessPath;
    }



    public int getChildPositionCode(ProgramElement child) {
        // role 0: prefix
        // role 1 (IDX): parameters
        if (prefix == child) {
            return 0;
        }
        if (arguments != null) {
            int index = getArrayPos(arguments,child);
            if (index >= 0) {
                return (index << 4) | 1;
            }
        }
        return -1;
    }

    /**
     *      Get reference prefix.
     *      @return the reference prefix.
     */
    public ReferencePrefix getReferencePrefix() {
        return prefix;
    }

    public SourceElement getFirstElement() {
        return (prefix == null) ? this : prefix.getFirstElement();
    }

    /** calls the corresponding method of a visitor in order to
     * perform some action/transformation on this element
     * @param v the Visitor
     */
    public void visit(Visitor v) {
	v.performActionOnSuperConstructorReference(this);
    }

    public void prettyPrint(PrettyPrinter p) throws java.io.IOException {
        p.printSuperConstructorReference(this);
    }
}
