// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

/* Generated by Together */
/** creates a simple textual representation of the formula it visits */ 

package de.uka.ilkd.key.logic;

import java.util.Vector;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.op.*;


public class ProofObligationPrinter extends Visitor {
    /** empty StringBuffer, ATTENTION: Take care that you'll never
     * change this buffer */
    public static final StringBuffer EMPTY_STRINGBUFFER=new StringBuffer();

    /** to store some signs in this way, may be helpful if we want to
     * extend this class for latex style etc. */ 
    protected StringBuffer LPAREN=new StringBuffer("(");
    protected StringBuffer RPAREN=new StringBuffer(")");
    protected StringBuffer LBRACK=new StringBuffer("[");
    protected StringBuffer RBRACK=new StringBuffer("]");
    protected StringBuffer LT=new StringBuffer("<");
    protected StringBuffer GT=new StringBuffer(">");
    protected StringBuffer COMMA=new StringBuffer(",");
    protected StringBuffer SPACE=new StringBuffer(" ");
    protected StringBuffer SORTSEP=new StringBuffer(".");

    private Term visited;
    private Vector vector;
    private int currentLevel;

    public ProofObligationPrinter() {
	vector=new Vector();
	currentLevel = -1;
    }

    /** removes the last n elements of the vector */
    private void removeLast(int n) {
	int last=vector.size()-1;
	for (int i=0;i<n;i++) {
	    vector.remove(last);
	    last--;
	}
    }

    private StringBuffer print(Operator op, int arity) {
	if (op instanceof Quantifier) {
	    return print((Quantifier)op, arity);
	} else if (op instanceof Modality) {
	    return print((Modality)op, arity);
	} else if (op instanceof Op) {
	    return print((Op)op, arity);
	} 	
	return print((TermSymbol)op, arity);		
    }    

    // prints an operator term
    private StringBuffer print(Op operator, int arity) {
	StringBuffer result=new StringBuffer();
	StringBuffer op=new StringBuffer(operator.name().toString());	
	int last=vector.size()-1;

	if (arity==1) { //prefix operator
	    result.append(op);
	    result.append(SPACE);
	    result.append((StringBuffer)vector.get(last));
	} else { // s.th. like "(P(x) and Q(x,y))"	  
	    result.append(LPAREN);
	    result.append((StringBuffer)vector.get(last-arity+1));
	    for (int i=1;i<arity;i++) {  
		result.append(SPACE);
		result.append(op);
		if(operator == Op.OR) {
		    result.append("\n"); // go to newline ...
		}
		result.append(SPACE);
		result.append((StringBuffer)vector.get(last-arity+1+i));		
	    }
	    result.append(RPAREN);
	} 
	
	removeLast(arity);
	return result;
    }


    
    // prints a diamond term
    private StringBuffer print(Modality mod, int arity) {
	StringBuffer result = new StringBuffer();
	if (mod == Op.DIA) { 
	    result.append(LT);
	} else { // box operator
	    result.append(LBRACK);
	}
	result.append((((ProgramTerm)visited).javaBlock()).toString());
	if (mod == Op.DIA) {
	    result.append(GT);
	} else { // box operator
	    result.append(RBRACK);
	}
	result.append(LPAREN);
	result.append((StringBuffer)vector.remove(vector.size()-1));
	result.append(RPAREN);	   	
	return result;
    }

    // prints a proofobl term
    private StringBuffer printProofOblTerm(ProofObligationTerm p) {
	StringBuffer result=new StringBuffer();
	result.append(p.op().name());
	result.append(LPAREN);
	result.append(p.ruleVar().name());
	result.append(RPAREN);	   	
	return result;
    }

    // prints a quantifier term
    private StringBuffer print(Quantifier quantifier, int arity) {
	StringBuffer result=new StringBuffer();
	result.append(quantifier.name());
	// print bound variable list
	result.append(LBRACK);
	ArrayOfQuantifiableVariable vars
	    = ((QuantifierTerm)visited).varsBoundHere(0);
	for (int i=0;i<vars.size()-1;i++) {
	    result.append(vars.getQuantifiableVariable(i).name());
	    //result.append(SORTSEP);
	    //result.append(vars.getQuantifiableVariable(i).sort());
	    result.append(COMMA);
	}
	result.append(vars.getQuantifiableVariable(vars.size()-1).name());
	//result.append(SORTSEP);
	//result.append(vars.getQuantifiableVariable(vars.length-1).sort());

	result.append(RBRACK);

	// print subterm
	result.append((StringBuffer)vector.get
		      (vector.size()-1));
	
	removeLast(1);
	return result; 
    }

    // prints a function or predicate term
    private StringBuffer print(TermSymbol termSym,int arity) {
	// no infix output ...
	StringBuffer result = null;
	if (arity!=2 ||( (termSym.sort()==Sort.FORMULA) )) {
	    result=new StringBuffer(termSym.name().toString());
	int last=vector.size();
	if (arity!=0) {
	    result.append(LPAREN);
	    result.append((StringBuffer)vector.get(last-arity));
	    for (int i=1;i<arity;i++) {
		result.append(COMMA);
		result.append((StringBuffer)vector.get(last-arity+i));
	    }	        
	    result.append(RPAREN);
	}
	removeLast(arity);
	} else {
	    // use infix output ...
	    result=new StringBuffer();
	    int last=vector.size();
	
	    result.append(LPAREN);
	    result.append((StringBuffer)vector.get(last-2));
	    result.append(" " + termSym.name().toString() + " ");
	
	    result.append((StringBuffer)vector.get(last-1));
	    result.append(RPAREN);
	
	removeLast(arity);
	}
	return result;
    }

    public void visit(Term visited) {
	this.visited=visited;
	int arity=visited.arity();
	StringBuffer term=EMPTY_STRINGBUFFER;
	
	if (!(visited instanceof ProofObligationTerm)) {	
	    term=print(visited.op(),arity); 	
	} else {
	    term = printProofOblTerm((ProofObligationTerm) visited);
	}

	vector.add(term);
    }
    
    public void subtreeEntered(Term subtreeRoot){
	currentLevel++;
	
    }
public void subtreeLeft(Term subtreeRoot){
	currentLevel--;
	
    }

    public String toString() {
	if (vector.isEmpty()) {
	    return "";
	}
	return ((StringBuffer)vector.remove(0)).toString(); 	
    }
}
