// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

/* Generated by Together */

package de.uka.ilkd.key.logic;

import de.uka.ilkd.key.proof.SkolemDeliverer;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.rule.Rulevariable;

import java.util.*;

/**
 * visitor for <t> execPostOrder </t> of 
 * {@link de.uka.ilkd.key.logic.Term}. Called with that method
 * on a term, the visitor skolemizes all occurences of proofobl(r) terms
 * (where r is a rulevariable) by a constant skolempredicate.
 * The skolempredicates for proofobl term for the same rulevariable r should be the same.
 */

public class ProofObligationTermSkolemizer extends Visitor {
    
    /** maps a rulevariable to the corresponding constant skolem predicate */ 

    private Map ruleVarsMap;

    /** the skolemdeliverer to be used for the skolemization */
    private SkolemDeliverer skolemDeliverer;

    /**
     * the stack contains the subterms that will be used in getTerm() in Term
     * in order to build the new term. A boolean value
     * between or under the subterms on the stack indicate that a term using
     * these subterms should build a new term instead of using the old one,
     * because one of its subterms has been built, too.
     */
    private Stack subStack; //of Term (and Boolean)
    private TermFactory tf=TermFactory.DEFAULT;
    private Boolean newMarker=new Boolean(true);

   

    public  ProofObligationTermSkolemizer(SkolemDeliverer sd) {
	this.skolemDeliverer = sd;
	this.ruleVarsMap = new HashMap();
	subStack=new Stack(); // of Term
	
    }

    private Term[] neededSubs(int n) {
	boolean newTerm=false;
	Term[] result=new Term[n];
	for (int i=n-1; i>=0; i--) {
	    Object top=subStack.pop();
	    if (top==newMarker){
		newTerm=true; 
		top=subStack.pop();
	    }
	    result[i]=(Term) top;
	}
	if (newTerm) subStack.push(newMarker);
	return result;
    }

    public void visit(Term visited) {
	// Sort equality has to be ensured before calling this method
	ProofObligationTerm p;
	Rulevariable rv;
	Term replaceTerm;

	// vars bound directly by operator of visited term ...
	ArrayOfQuantifiableVariable vars =(visited.varsBoundHere(0).size() == 0)
	    ? visited.varsBoundHere(1) : visited.varsBoundHere(0);
	if (visited instanceof ProofObligationTerm) {
	    p = (ProofObligationTerm)visited;
	    

	    rv = p.ruleVar();
	    
	    if (ruleVarsMap.containsKey(rv)){
		replaceTerm = (Term)ruleVarsMap.get(rv); 
	    } else { // create a new constant skolempredicate ...
		SkolemFunction sf = skolemDeliverer.getSkolemSymbolForSort(Sort.FORMULA,
									   new PrimitiveSort[0]);
		replaceTerm = tf.createSkolemFunctionTerm(sf, new Term[0]);
		// save term in mapping ...
		ruleVarsMap.put(rv,replaceTerm);
	    }
	    
	    //pop result on stack and mark as new ...
	    subStack.push(replaceTerm);
	    subStack.push(newMarker);
	    
	    
	   	  
	    
	} else {	 
	    Term[] neededsubs=neededSubs(visited.arity());
	    if (!subStack.empty() && subStack.peek()==newMarker) {
		subStack.pop(); // delete new marker ...
		subStack.push(tf.createTerm(visited.op(), 
					    neededsubs,
					    vars,
					    visited.javaBlock(),
					    visited.ruleVar()));
		subStack.push(newMarker); // add new marker ...
		
	    } else {
		subStack.push(visited);
	    }
	    
	    
	    
	}
    }      
    

    
    /**
     * delivers the new built term
     */
    public Term getTerm() {
	Object result = subStack.pop();
	if (result == newMarker) {
	    result = subStack.pop();
	}
	return((Term)result);
    }
    


}









