// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.logic;

import de.uka.ilkd.key.logic.op.Operator;
import de.uka.ilkd.key.logic.op.QuantifiableVariable;
import de.uka.ilkd.key.logic.op.ArrayOfQuantifiableVariable;
/**
 * A SubstitutionTerm is an object that contains an Operator two subterms. 
 * and one variable
 * Instances should never be accessed via 
 * this interface, use the interface of the superclass Term and construct 
 * instances only via
 * a TermFactory instead.
 */

class SubstitutionTerm extends Term {

    /** variable to substitute
     * @link aggregation
     * @supplierCardinality * 
     */
    private final ArrayOfQuantifiableVariable substVar;
    
    /** term being changed */
    private final Term origTerm;

    /** term replacing the variable substVar*/
    private final Term substTerm;
    
    /** depth of the term */
    private final int depth;
    

    /** creates a substitution term
     * @param substVar the Variable to be substituted
     * @param substTerm the Term that replaces substVar
     * @param origTerm the Term that is substituted
     */
    public SubstitutionTerm(Operator op, QuantifiableVariable substVar, 
			    Term substTerm, Term origTerm) {
	super(op);
	this.substTerm = substTerm;
	this.origTerm  = origTerm;
	this.substVar  = new ArrayOfQuantifiableVariable
	    (new QuantifiableVariable[] {substVar});

	depth = 1+(substTerm.depth() > origTerm.depth() ? 
		   substTerm.depth() :  origTerm.depth());

	setSort();
	determineFreeVars();
    }

    /** @return n-th subterm ( e.g. {x s} t (0:s ,  1: t) */    
    public Term sub(int n) {
	return (n==0 ? substTerm : origTerm);
    }	

    /**@return depth of the term */
    public int depth() {
	return depth;
    }
   
    /** @return arity of the substitution term 1 as int */
    public int arity() {
	return 2;
    } 

    /** @return the bound variables of the n-th subterm */
    public ArrayOfQuantifiableVariable varsBoundHere(int n) {
	return (n==0 ? EMPTY_VAR_LIST : substVar);
    }

}
