// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.logic;

import junit.framework.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.abstraction.PrimitiveType;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.logic.op.Function;
import de.uka.ilkd.key.logic.op.ProgramVariable;
import de.uka.ilkd.key.logic.op.*;

public class TestUpdateSimplifier extends TestCase {


    Function a;
    Function b;
    Function c;
    de.uka.ilkd.key.java.declaration.FieldSpecification vsp_x;
    de.uka.ilkd.key.java.declaration.FieldSpecification vsp_y;
    Sort integer;
    ProgramVariable x;
    ProgramVariable y;
    Function const2;
    Term asub1;
    Term asub2;
    de.uka.ilkd.key.java.Recoder2KeY c2k ;
	


    public TestUpdateSimplifier(String s) {
	super(s);
    }

    public void setUp() {
	JavaInfo ji=new JavaInfo();
	c2k = new de.uka.ilkd.key.java.Recoder2KeY(ji.rec2key(),
						   ji.getTypeConverter()); 
	PrimitiveType primInt = PrimitiveType.JAVA_INT; 
	Sort integer = new PrimitiveSort(new Name("integer"));
	KeYJavaType kjt=new KeYJavaType(primInt, integer);

 	x = new ProgramVariable(new ProgramElementName("x"), kjt);

 	y = new ProgramVariable(new ProgramElementName("y"), kjt);
	vsp_x = new de.uka.ilkd.key.java.declaration.FieldSpecification
 		(x, null, kjt);
 	vsp_y = new de.uka.ilkd.key.java.declaration.FieldSpecification
 		(y, null, kjt);

	const2=new Function(new Name("2"), integer, new PrimitiveSort[0]);
	a=new Function(new Name("A"),Sort.FORMULA, new Sort[] {integer, integer});
	c=new Function(new Name("C"),integer, new Sort[] {integer, integer});
	b=new Function(new Name("B"),Sort.FORMULA, new PrimitiveSort[0]);
	asub1 = TermFactory.DEFAULT.createFunctionTerm(x);
	asub2 = TermFactory.DEFAULT.createFunctionTerm(y);
    }
    
    public void testUpdates() {

	Term aorb=TermFactory.DEFAULT.createJunctorTerm
	    (Op.OR, 
	     TermFactory.DEFAULT.createUpdateTerm
	     (y, TermFactory.DEFAULT.createFunctionTerm(const2), 
	      TermFactory.DEFAULT.createFunctionTerm(a, asub1, asub2 )), 
	      TermFactory.DEFAULT.createFunctionTerm(b));
	Term t=TermFactory.DEFAULT.createUpdateTerm
	    (x, TermFactory.DEFAULT.createFunctionTerm(const2), aorb); 
	UpdateSimplifier simplifier=new UpdateSimplifier();
	Term result = simplifier.simplify(t);
	assert(result.sub(0).sub(0).op()==const2);
	assert(result.sub(0).sub(1).op()==const2);
    }


    public void testUpdates2() {
	TermFactory tf=TermFactory.DEFAULT;
	// a(x,y)
	Term aof = tf.createFunctionTerm(a, asub1, asub2 ); 
	// b
	Term xb=tf.createFunctionTerm(b);
	// <{}>b
	Term dia_b = tf.createDiamondTerm(JavaBlock.EMPTY_JAVABLOCK, xb);
	// (b and <{}>b)
	Term b_and_dia_b = tf.createJunctorTerm(Op.AND, xb, dia_b);
	// (b and <{}>b) or (a(x,y) imp <{}>a(x,y)) 	
	Term aorb=TermFactory.DEFAULT.createJunctorTerm
	    (Op.OR, b_and_dia_b,
	     tf.createJunctorTerm(Op.IMP, aof,
				  tf.createDiamondTerm(JavaBlock.EMPTY_JAVABLOCK, aof)));
	//  {y:=const2} (b and <{}>b) or (a(x,y) imp <{}>a(x,y)) 
	Term t=//tf.createUpdateTerm
	    //	    (x, 
	    //tf.createFunctionTerm(const2),
		    tf.createUpdateTerm(y,// tf.createFunctionTerm(c,asub1,asub1)
			tf.createFunctionTerm(const2),aorb);//); 
	UpdateSimplifier simplifier=new UpdateSimplifier();
	Term result = simplifier.simplify(t); 
	// expected result (b and {y:=const2}<{}>b) or (a(x,const2) imp {y:=const2}<{}>a(x,y)
	assert("First subterm of >>or<< should be (B and <{}>B) "+
	       " but is "+result.sub(0), result.sub(0).op() == Op.AND &&
	       result.sub(0).sub(0).op() == b &&  
	       result.sub(0).sub(1) == dia_b);	

	assert("Second subterm of >>or<< should have top operator imp but has "+
	       result.sub(1).op(), result.sub(1).op() == Op.IMP); 

	assert("First subterm of >>imp<< should be a(x,2) "+
	       " but is "+result.sub(1).sub(0), 
	       result.sub(1).sub(0).op() == a &&
	       result.sub(1).sub(0).sub(0).op()==x &&
	       result.sub(1).sub(0).sub(1).op()==const2);

	assert("Second subterm of >>imp<< should be {y:=2}<{}>a(x,y) "+
	       " but is "+result.sub(1).sub(1), 
	       result.sub(1).sub(1).op() instanceof UpdateOp &&
	       ((UpdateOp)result.sub(1).sub(1).op()).variable().op() == y &&
	       result.sub(1).sub(1).sub(0).op() == const2 &&
	       result.sub(1).sub(1).sub(1).op() == Op.DIA &&
	       result.sub(1).sub(1).sub(1).sub(0) == aof);	
    }


    public void testUpdates3() {
	TermFactory tf=TermFactory.DEFAULT;
	// a(x,y)
	Term aof = tf.createFunctionTerm(a, asub1, asub2 ); 
	// b
	Term xb=tf.createFunctionTerm(b);
	// <{}>b
	Term dia_a2 = tf.createDiamondTerm(JavaBlock.EMPTY_JAVABLOCK, aof);
	// (b and <{}>b)
	Term b_and_dia_a2 = tf.createJunctorTerm(Op.AND, xb, dia_a2);
	// (b and <{}>b) or (a(x,y) imp <{}>a(x,y)) 	
	Term aorb=TermFactory.DEFAULT.createJunctorTerm
	    (Op.OR, b_and_dia_a2,
	     tf.createJunctorTerm(Op.IMP, aof,
				  tf.createDiamondTerm(JavaBlock.EMPTY_JAVABLOCK, aof)));
	//  {y:=const2} (b and <{}>b) or (a(x,y) imp <{}>a(x,y)) 
	Term t2=//tf.createUpdateTerm
	    //	    (x, 
	    //tf.createFunctionTerm(const2),
		    tf.createUpdateTerm(y,// tf.createFunctionTerm(c,asub1,asub1)
			tf.createFunctionTerm(const2),aorb);//); 
	//  {x:=const2}{y:=const2} (b and <{}>b) or (a(x,y) imp <{}>a(x,y)) 
	Term t=//tf.createUpdateTerm
	    //	    (x, 
	    //tf.createFunctionTerm(const2),
		    tf.createUpdateTerm(x,// tf.createFunctionTerm(c,asub1,asub1)
			tf.createFunctionTerm(const2),t2);//); 

	UpdateSimplifier simplifier=new UpdateSimplifier();
	Term result = simplifier.simplify(t); 
	// expected result (b and {y:=2}<{}>b) or (a(2,2) imp {y:=2}<{}>a(x,y)
	assert("First subterm of >>or<< should be (B and {x:=2}{y:=2}<{}>B) "+
	       " but is "+result.sub(0), result.sub(0).op() == Op.AND &&
	       result.sub(0).sub(0).op() == b && 
	       result.sub(0).sub(1).op() instanceof UpdateOp &&
	       result.sub(0).sub(1).sub(0).op() == const2 &&  
	       result.sub(0).sub(1).sub(1).sub(0).op() == const2 &&  
	       result.sub(0).sub(1).sub(1).sub(1) == dia_a2);	

	assert("Second subterm of >>or<< should have top operator imp but has "+
	       result.sub(1).op(), result.sub(1).op() == Op.IMP); 

	assert("First subterm of >>imp<< should be a(2,2) "+
	       " but is "+result.sub(1).sub(0), 
	       result.sub(1).sub(0).op() == a &&
	       result.sub(1).sub(0).sub(0).op()==const2 &&
	       result.sub(1).sub(0).sub(1).op()==const2);

	assert("Second subterm of >>imp<< should be {x:=2}{y:=2}<{}>a(x,y) "+
	       " but is "+result.sub(1).sub(1), 
	       result.sub(1).sub(1).op() instanceof UpdateOp &&
	       ((UpdateOp)result.sub(1).sub(1).op()).variable().op() == x &&
	       result.sub(1).sub(1).sub(0).op() == const2 &&
	       result.sub(1).sub(1).sub(1).sub(0).op() == const2 &&
	       result.sub(1).sub(1).sub(1).sub(1).op() == Op.DIA &&
	       result.sub(1).sub(1).sub(1).sub(1).sub(0) == aof);	
    }

    
    public void testUpdatesStopAtModality() {

	//  (({y:=2}a(x,y)) or b )
	Term aorb=TermFactory.DEFAULT.createJunctorTerm
	    (Op.OR, TermFactory.DEFAULT.createUpdateTerm
	     (y, TermFactory.DEFAULT.createFunctionTerm(const2), 
	      TermFactory.DEFAULT.createFunctionTerm(a, asub1, asub2 )), 
	     TermFactory.DEFAULT.createFunctionTerm(b));

	//  (<{}>({y:=2}a(x,y)) or b )	
	Term modTerm = TermFactory.DEFAULT.createDiamondTerm
	    (c2k.readBlockWithEmptyContext(" {} "), aorb);
	//  ({x:=2}<{}>(({y:=2}a(x,y)) or b) )	
	Term t=TermFactory.DEFAULT.createUpdateTerm
	    (x, TermFactory.DEFAULT.createFunctionTerm(const2), modTerm); 
	UpdateSimplifier simplifier=new UpdateSimplifier();
	Term result = simplifier.simplify(t);       
	// expected result:({x:=2}<{}>((a(x,2)) or b) )	
	assert("top operator should be an update operator but is "+result.op(),
	       result.op() instanceof UpdateOp); 
	assert("Diamond expected but is "+result.sub(1).op()
	       ,result.sub(1).op() == Op.DIA);
	assert(">>or<< expected but is "+result.sub(1).sub(0).op(),
	       result.sub(1).sub(0).op() == Op.OR);
	assert(">>a<< expected but is "+result.sub(1).sub(0).sub(0).op(),
	       result.sub(1).sub(0).sub(0).op() == a);
	assert("2 expected but is "+result.sub(1).sub(0).sub(0).sub(1).op(),
	       result.sub(1).sub(0).sub(0).sub(1).op() == const2);
	}
}
