// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001 Universitaet Karlsruhe, Germany
//                    and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General 
// Public License. See LICENSE.TXT for details.
//

package de.uka.ilkd.key.logic.op;
import de.uka.ilkd.key.java.abstraction.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.Expression;
import de.uka.ilkd.key.java.Statement;
import de.uka.ilkd.key.java.StatementBlock;
import de.uka.ilkd.key.java.ArrayOfStatement;
import de.uka.ilkd.key.java.ArrayOfExpression;
import de.uka.ilkd.key.java.expression.operator.CopyAssignment;
import de.uka.ilkd.key.java.statement.MethodCallStatement;
import de.uka.ilkd.key.java.reference.*;
import de.uka.ilkd.key.java.reference.MethodReference;
import de.uka.ilkd.key.java.declaration.Modifier;
import de.uka.ilkd.key.java.declaration.MethodDeclaration;
import de.uka.ilkd.key.java.declaration.LocalVariableDeclaration;
import de.uka.ilkd.key.java.declaration.VariableSpecification;
import de.uka.ilkd.key.java.declaration.ParameterDeclaration;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;

import de.uka.ilkd.key.java.statement.*;
import de.uka.ilkd.key.java.expression.operator.*; 
import de.uka.ilkd.key.java.expression.literal.*;
import de.uka.ilkd.key.rule.metaconstruct.ProgramMetaConstruct;
import java.util.Vector;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.rule.inst.SVInstantiations;
 



/** 
 * Symbolically executes a method invocation
 */ 
public class MethodCall extends ProgramMetaConstruct {
    
    private final SchemaVariable result;

    /** creates the methodcall-MetaConstruct 
     * @param result the SchemaVariable that is used to keep the result
     * @param body the ProgramElement contained by the meta construct 
     */
    public MethodCall(SchemaVariable result, ProgramElement body) {
	super(new Name("method-call"), body);
	this.result = result;
    }

    /** creates the methodcall-MetaConstruct 
     * @param body the ProgramElement contained by the meta construct 
     */
    public MethodCall(ProgramElement body) {
	super(new Name("method-call"), body);
	this.result = null;
    }


    /** gets an array of expression and returns a list of types */
    private ListOfKeYJavaType getTypes(JavaInfo javaInfo, 
				       ArrayOfExpression args) {
	ListOfKeYJavaType result = SLListOfKeYJavaType.EMPTY_LIST; 
	for (int i = args.size()-1; i >= 0 ; i--) {
	    Expression argument = args.getExpression(i);
	    result = result.prepend
		(javaInfo.getTypeConverter().getPrimitiveKeYJavaType(argument));
	}
	return result;
    }



   private MethodReference methRef;
   private ProgramMethod pm;
   private ReferencePrefix refPrefix;
   private ReferencePrefix newContext;

    /** performs the program transformation needed for symbolic
     * program execution 
     * @param javaInfo the JavaInfo with all necessary information 
     * about the java programs
     * @param svInst the instantiations esp. of the inner and outer label 
     * @return the transformed program
     */
    public ProgramElement symbolicExecution(ProgramElement pe,
					    JavaInfo javaInfo,
					    SVInstantiations svInst) {

	ProgramVariable pvar = null;
	if (result != null) {
	    pvar = (ProgramVariable) svInst.getInstantiation(result);
	}

	methRef = (MethodReference) pe;
	refPrefix = methRef.getReferencePrefix();
	StatementBlock ourBody;

	
        Type t = null;
	pm = methRef.method(javaInfo);
        newContext = methRef.getReferencePrefix();

	if (refPrefix instanceof TypeRef) {
	    t = javaInfo.getKeYJavaType
		(((TypeRef)refPrefix).getName());		
	    if (t == null) {
System.err.println("CHECK: t==null");            
		t = javaInfo.getKeYJavaTypeByClassName
		    (((TypeRef)refPrefix).getName()).getJavaType();
	    }
	} else if (refPrefix instanceof ProgramVariable) {
           t = ((ProgramVariable)refPrefix).getKeYJavaType();
	} else if (refPrefix instanceof FieldReference) {
           t = ((FieldReference)refPrefix).getProgramVariable().getKeYJavaType();
	} else {
	    throw new de.uka.ilkd.key.util.NotSupported
		("Unsupported method invocation mode\n"+
		 refPrefix.getClass());
	}			    


/*
	if (pm == null) {
	   System.err.println("Failed to get method by .method()");
	   pm = javaInfo.getProgramMethod(
		                (KeYJavaType) t, methRef.getName(), 
		                getTypes(javaInfo, methRef.getArguments()));
	}
	if (pm == null) throw new RuntimeException("No method found for "+t);
*/


        de.uka.ilkd.key.casetool.together.ModelUtil.hiliteMethod(t, pm);


// Static invocation mode
	if (pm.isStatic()) {

//            newContext = new TypeRef((KeYJavaType)t);
// this was too aggressive
// - we did not transform the logical formula
// - what about side effects in prefix? (OK, we should take care of them 
//  beforehand, but anyway...)

	    ourBody = new StatementBlock(
	                 new MethodBodyStatement(t, methRef,  newContext));

// Instance invocation mode
	} else  {
	   Type javaType = ((KeYJavaType)t).getJavaType();
	   if (pm.isPrivate()) { // private methods are bound statically
	      ourBody = new StatementBlock(
	                   new MethodBodyStatement(javaType, methRef,
						   newContext));

	   } else {
              ifCascadeTop = ifCascadeBottom = null;
	      javaInfo.getKeYProgModelInfo().findImplementations(this, 
                 (KeYJavaType)t, methRef.getName(),
                 getTypes(javaInfo, methRef.getArguments()));
	      ourBody = new StatementBlock(ifCascadeTop);

           }
        }
        return new MethodCallStatement(pvar, ourBody, methRef);
    }



//***************** Dynamic Binding Construction Utilities ***************

   private If ifCascadeTop = null;
   private If ifCascadeBottom = null;


   private MethodBodyStatement makeMbs(Type t) {
      if (t instanceof KeYJavaType) 
	 t = ((KeYJavaType)t).getJavaType();
      return new MethodBodyStatement(t, methRef, newContext);
   }


   public Expression makeIOf(Type t) {
      return new Instanceof((Expression) refPrefix, 
                            new TypeRef((KeYJavaType)t));
   }


   public Expression makeEIOf(Type t) {
      return new ExactInstanceof((ProgramVariable) refPrefix, 
                            new TypeRef((KeYJavaType)t));
   }



   public Expression disj(Vector v) {
      v.remove(null);
      if (v.size() == 0) return null;
      Expression e = makeIOf((Type)v.get(0));
      for (int i=1; i< v.size(); i++) 
        e = new LogicalOr(e, makeIOf((Type)v.get(i)));
      return e;
   }


   public void addBranch(Expression ifCond, Type t) {
      if (ifCascadeTop == null) 
         ifCascadeBottom = ifCascadeTop = new If(ifCond, new Then(makeMbs(t)));
      else
         ifCascadeBottom.addBranch(
            new Else(ifCascadeBottom = new If(ifCond, new Then(makeMbs(t)))));
   }


}
