// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.logic.op;

import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Term;
/** this class is used to represent a dynamic logic modality like
 * diamond and box (but also extensions of DL like
 * preserves and throughout are possible in the future).
 * For further information see @see PrpgramTerm.
 */

public class Modality extends Op {
    
    /** creates a modal operator with the given name
     * @param name the Name of the modality 
     */
    Modality(Name name) {
	super(name);
    }

    /** @return Sort.FORMULA
     */
     public Sort sort(Term term) {
        return Sort.FORMULA;
    }  

    /** @return true if the subterm at postion 0 of the given term 
     * has Sort.FORMULA and the arity of the term is 1.
     */
    public boolean validTopLevel(Term term){
	if (term.arity()!=1) return false;
        return term.sub(0).sort().equals(Sort.FORMULA);
    }

    /** @return arity of the Diamond operator as int */
    public int arity() {
	return 1;
    }


}
