// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.ocl.tudresden;

import tudresden.ocl.check.types.*;
import tudresden.ocl.check.OclTypeException;

import com.togethersoft.openapi.sci.*;
import com.togethersoft.openapi.sci.enum.*;

import java.util.HashMap;

import recoder.service.*;
import recoder.abstraction.Package;
import recoder.list.*;
import recoder.convenience.*;
import recoder.io.*;
import recoder.util.*;
import recoder.java.*;

public class JavaOCLTypeMap{

    private Type getTypeFor(String t) {
	if (t.equals("byte") || t.equals("short") || 
	    t.equals("char") || t.equals("int")  || 
	    t.equals("long")) {
	    return Basic.INTEGER;
	} else if (t.equals("float") || t.equals("double")) {
	    return Basic.REAL;
	} else if (t.equals("boolean")) {
	    return Basic.BOOLEAN;
	} else if (t.equals("void")) {
	    return Any.VOID;
	} else
	    throw new OclTypeException("Cannot find OCL type"+
				       "for Java type \""+t+"\"");
    }
    
    
    public Type getTypeFor(SciType type, HashMap classes){
	if (isBasicType(type)){
	    // basic type
	    return getTypeFor(type.getCanonicalText());
	} else if (type.getCanonicalText().equals("java.lang.String")){
	    //System.out.println("String found: " + type.getReferencedElement().getQualifiedName());
	    //System.out.println("Ist aber    : " + type.getCanonicalText());
	    return Basic.STRING;
	} else {
	    String refClass = type.getReferencedElement().getName();
	    if (classes.containsKey(refClass) && refClass!=null) {
		return (Type)classes.get(type.getReferencedElement().getName());
	    }
	    else {
		// 		throw new OclTypeException("Cannot find OCL type "+
		// 					   "for \""+refClass+"\"");
		//		System.out.println("ATTENTION! Did not find OCL type for \""+
		//				   refClass+"\". Instead of giving up I assume that \""+
		//				   refClass+"\" is a class from a stanard java library "+
		//				   "and I give it the OCL type ANY.");
		//		System.out.println("Genauer: Typ " + type.getCanonicalText());
		//		return Basic.OCLANY;
		
		//		System.out.println("Created incomplete TJClassifier: " + type.getCanonicalText());
		return new TJClassifier(type.getCanonicalText(),
					type.getCanonicalText(), new HashMap());
	    }
	    
	}	
    }
    
    private boolean isBasicType(SciType type){
	if (type.getReferencedElement()==null)
	    return true;
	else 
	    return false;
    }
    
    private boolean isBasicType(recoder.abstraction.Type type){
	if ((type instanceof recoder.abstraction.PrimitiveType)
	    | (type instanceof recoder.abstraction.NullType)
            | (type == null)) {
			return true;
	} else  {
	    return false;
		}
    }
    
    public Type getTypeFor(recoder.abstraction.Type type, HashMap classes){
        String result;
        if (type == null) {
	    return getTypeFor("void");
        } else if (isBasicType(type)){
	    // basic type
	    return getTypeFor(type.getName());
	} else if (type.getFullName().equals("java.lang.String")){
	    return Basic.STRING;
	} else {
	    /*            try {
			  recoder.abstraction.TypeDeclaration td = (recoder.abstraction.TypeDeclaration)type;
			  CompilationUnit cu = sfr.getCompilationUnit(type.getFullName);
			  if (cu != null) {
			  
			  }
			  
			  } catch (Exception e) {
			  }
	    */
	    String refClass = type.getFullName();
	    if (classes.containsKey(refClass) && refClass!=null) {
		return (Type)classes.get(refClass);
	    }
	    else {
		return new TJClassifier(type.getFullName(), 
					type.getFullName(), new HashMap());
            }
        }
    }
    
}
