// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001 Universitaet Karlsruhe, Germany
//                    and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General 
// Public License. See LICENSE.TXT for details.
//
// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001 Universitaet Karlsruhe, Germany
//
// The KeY system is protected by the GNU General 
// Public License. See LICENSE.TXT for details.

/* -*-antlr-*- */
header {

    package de.uka.ilkd.key.parser;

    import java.io.*;
    import antlr.*;
    import de.uka.ilkd.key.logic.*;
    import de.uka.ilkd.key.rule.*;
    import de.uka.ilkd.key.logic.op.*;
    import de.uka.ilkd.key.logic.sort.*;
    import de.uka.ilkd.key.java.visitor.*;
    import java.util.Vector;
    import java.util.HashMap;
    import de.uka.ilkd.key.proof.UseInsteadMap;

}

/** 
 * A parser for signature declarations,
 * i.e. sorts, schema variabls, predicates, functions and heuristics.
 */  

class DeclParser extends Parser;
options {
	importVocab=KeYLexer;
	k = 2;
	defaultErrorHandler=false;
}

{
	static final Sort[] AN_ARRAY_OF_SORTS = new Sort[0];

	private Namespace var_ns;
	private Namespace func_ns;
	private Namespace sort_ns;
	private Namespace heur_ns;
	
	// these variables are set if a file is read in step by
	// step. This used when reading in LDTs because of cyclic
	// dependencies.
	protected boolean skip_schemavariables;
	protected boolean skip_functions;
	protected boolean skip_predicates;
	protected boolean skip_sorts;
	protected boolean skip_heuristics;
	
	protected UseInsteadMap useInstead=new UseInsteadMap();


	public DeclParser(TokenStream lexer,
			  String filename,
			  Namespace var_ns,
			  Namespace func_ns,
			  Namespace sort_ns,
			  Namespace heur_ns) {
	    this(lexer);
	    setFilename(filename);
	    this.var_ns = var_ns;
	    this.func_ns = func_ns;
	    this.sort_ns = sort_ns;
	    this.heur_ns = heur_ns;

	    setSkips(false, false, false, false, false);
	}
		
	public Namespace getSorts() {
	    return sort_ns;
	}

	public Namespace functions() {
	    return func_ns;
	}

	public Namespace heuristics() {
	    return heur_ns;
	}

	public Namespace variables() {
	    return var_ns;
	}


	protected int getLine() {
	    int line = -1;
	    try {
		line = LT(0).getLine();
	    } 
	    catch (Exception e) {
		System.err.println("No further token in stream");
	    }
	    return line;
	}   

	protected int getColumn() {
	    int col = -1;
	    try {
		col = LT(0).getColumn();
	    } 
	    catch (Exception e) {
		System.err.println("No further token in stream");
	    }
	    return col;
	}   


        protected void setSkips(boolean svars, boolean funcs, 
				boolean preds, boolean sorts,
				boolean heurs) {
	    skip_schemavariables = svars;
   	    skip_functions  	 = funcs;
	    skip_predicates 	 = preds;
	    skip_sorts 		 = sorts;
	    skip_heuristics 	 = heurs;
	}

	public void parseVariables() 	
	    throws antlr.RecognitionException, antlr.TokenStreamException
	    {
		setSkips(false, true, true, true, true); 
		top();
		setSkips(false, false, false, false, false);	    		
	    }

	public void parseFunctions()
	    throws antlr.RecognitionException, antlr.TokenStreamException

	    {
		setSkips(true, false, true, true, true); 
		top();
		setSkips(false, false, false, false, false);
	    }

	public void parsePredicates()
	    throws antlr.RecognitionException, antlr.TokenStreamException
	    {
		setSkips(true, true, false, true, true); 
		top();
		setSkips(false, false, false, false, false);	    
	    }

	public void parseSorts() 
	    throws antlr.RecognitionException, antlr.TokenStreamException
	    {
		setSkips(true, true, true, false, true); 
		top();
		setSkips(false, false, false, false, false);
	    }

	public void parseHeuristics()
	    throws antlr.RecognitionException, antlr.TokenStreamException
	    {
		setSkips(true, true, true, true, false); 
		top();
		setSkips(false, false, false, false, false);	    
	    }

	public void parseFuncAndPred()
	    throws antlr.RecognitionException, antlr.TokenStreamException
	    {
		setSkips(true, false, false, true, true); 
		top();
		setSkips(false, false, false, false, false);
	    }

    public NamespaceSet namespaces() {
	NamespaceSet ns = new NamespaceSet();
	ns.setVariables(variables());
	ns.setFunctions(functions());
	ns.setSorts(getSorts());
	ns.setHeuristics(heuristics());
	return ns;
    }

}

top : decls;

decls :
		(
			sort_decls
		|
			schema_var_decls
		|
			pred_decls
		|
			func_decls
		|
			heur_decls
		) *
	;

sort_decls {
	boolean isObjectSort=false;
}:
		"sorts" 
		LBRACE
		(
			("object" {isObjectSort=true;})? id:IDENT SEMI
			{
			    Name sort_name = new Name(id.getText());
			    if (sort_ns.lookup(sort_name) == null) {
				Sort s;
				if (!isObjectSort) {
					s = new PrimitiveSort(sort_name);
				} else {
   			 	     if ("Object".equals(sort_name)) {
				         if (sort_ns.lookup(new Name("Object"))
								      ==null) {

						 s = new ObjectSort(sort_name);
					  }
					  s=(Sort)sort_ns.lookup(new Name("Object"));
					} else {
						s = new ObjectSort(sort_name, 
   					                 (Sort)sort_ns.lookup
                                                         (new Name("Object")));
					}	
				}
				sort_ns.add(s);
				if (s instanceof NonCollectionSort) {
					NonCollectionSort ns
						= (NonCollectionSort)s;
					sort_ns.add(ns.getSetSort());
					sort_ns.add(ns.getSequenceSort());
					sort_ns.add(ns.getBagSort());
				}
			    }
			}
			
		) *
		RBRACE
	;
		
schema_var_decls
{
	Sort s = null;
	boolean makeVariableSV = false;
}
	:
		"schema" "variables"
		LBRACE
		(
			{makeVariableSV = false;}
			(       ("program" 
					( "statement"
					{
						s=ProgramSVSort.STATEMENT;
					}
					| "statementlist"
					{
						s=ProgramSVSort.STATEMENTLIST;
					}
					| "expression"
					{
						s=ProgramSVSort.EXPRESSION;
					}
                                        | "expressionlist"
					{
						s=ProgramSVSort.EXPRESSIONLIST;
					}
					| "simple" "expression"
					{
						s=ProgramSVSort.SIMPLEEXPRESSION;
					}
					| "simple" "javaboolean" "expression"
					{
						s=ProgramSVSort.SIMPLEJAVABOOLEXPRESSION;
					}
					| "simple" "javabyte" "expression"
					{
						s=ProgramSVSort.SIMPLEJAVABYTEEXPRESSION;
					}
					| "simple" "javashort" "expression"
					{
						s=ProgramSVSort.SIMPLEJAVASHORTEXPRESSION;
					}
					| "simple" "javaint" "expression"
					{
						s=ProgramSVSort.SIMPLEJAVAINTEXPRESSION;
					}
					| "simple" "javalong" "expression"
					{
						s=ProgramSVSort.SIMPLEJAVALONGEXPRESSION;
					}
					| "simple" "javabyteshortint" "expression"
					{
						s=ProgramSVSort.SIMPLEJAVABYTESHORTINTEXPRESSION;
					}
					| "simple" "anyjavatype" "expression"
					{
						s=ProgramSVSort.SIMPLEANYJAVATYPEEXPRESSION;
					}
					| "nonsimple" "expression"	
					{
						s=ProgramSVSort.NONSIMPLEEXPRESSION;
					}
					| "javaboolean" "expression"	
					{
						s=ProgramSVSort.BOOLEANEXPRESSION;
					}
					| "nonsimple" "javaboolean" "expression"	
					{
						s=ProgramSVSort.NONSIMPLEBOOLEANEXPRESSION;
					} 
					| "lefthandside"
					{
						s=ProgramSVSort.LEFTHANDSIDE;
					}
					| "simple" "lefthandside"
					{
						s=ProgramSVSort.SIMPLELEFTHANDSIDE;
					}
					| "forinit"
					{
						s=ProgramSVSort.FORINIT;
					}
					| "statementexpressionlist"
					{
						s=ProgramSVSort.STATEMENTEXPRESSIONLIST;
					}
					| "label"
					{
						s=ProgramSVSort.LABEL;
					}
					| "type"
					{
						s=ProgramSVSort.TYPE;
					}
					| "variable"	
					{
						s=ProgramSVSort.VARIABLE;
					}
					| "variableinitializer"	
					{
						s=ProgramSVSort.VARIABLEINIT;
					}
					| "new"	
					{
						s=ProgramSVSort.NEW;
					}
					|		
					( "booleanvariable") {
						s=ProgramSVSort.BOOLEANVARIABLE;
					}
					|
					( "catchfinally") {
						s=ProgramSVSort.CATCHFINALLY;
					}
					|		
					( "methodreference") {
						s=ProgramSVSort.METHODREFERENCE;
					}
					|
					( "methodbody") {
						s=ProgramSVSort.METHODBODY;
					}
					|
					( "branchlist") {
						s=ProgramSVSort.BRANCHLIST;
					}
					)
				)
				|
				"formula"
				{ 
					s = Sort.FORMULA;
				}			
				|
				s=sort_id_check[true]
				( "variables" { makeVariableSV = true; } ) ?
				
			)
			one_schema_var_decl[s,makeVariableSV]
			( COMMA one_schema_var_decl[s,makeVariableSV] ) *
			SEMI
		) *
		RBRACE
	;

one_schema_var_decl[Sort s, boolean makeVariableSV] :
		id:IDENT
		{
		    if (!skip_schemavariables) {
			SchemaVariable v;
			if ( s == Sort.FORMULA ) {
				v = SchemaVariable.createFormulaSV(
						       new Name(id.getText())
						);
			} else if ( s instanceof ProgramSVSort ) {
				v = SchemaVariable.createProgramSV(
    						new Name(id.getText())
						,(ProgramSVSort) s
						);
			} else {
				if ( makeVariableSV ) {
 					v=SchemaVariable.createVariableSV(
					               new Name(id.getText()), s);
				} else {
					v = SchemaVariable.createTermSV(
						new Name(id.getText()), s);
				}
			}
			var_ns.add(v);
		    }
		}
	;






pred_decls 
{
	Sort[] argSorts;
	String pred_name;
}
	:
		"predicates"
		LBRACE
		(
			pred_name = id
			argSorts=arg_sorts[!skip_predicates]
			{
			    if (!skip_predicates) {
				Name predicate = new Name(pred_name);
				if (func_ns.lookup(predicate) != null) {
				    // throw new AmbigiousDeclException
				    //("predicate "+pred_name+" ", 
				    //getLine(), getColumn());
				} else {
				    Function p = new Function(predicate,
							      Sort.FORMULA, 
							      argSorts);
				    func_ns.add(p);			
				}
			    } 
			}
			SEMI
		) *
		RBRACE
	;

func_decls 
{
	Sort[] argSorts;
	Sort retSort;
        String func_name;
}
	:
		"functions"
		LBRACE
		(
			retSort = sort_id_check[!skip_functions]
			func_name = id
			argSorts = arg_sorts[!skip_functions]
			{
			    if (!skip_functions) {

				Name fct_name = new Name(func_name);
				if (func_ns.lookup(fct_name) != null) {
				    //throw new AmbigiousDeclException
				    //("function "+func_name, 
				    //getLine(), getColumn());
				} else {
				    Function f=new Function(fct_name,
							    retSort, 
							    argSorts);
				    func_ns.add(f);
				}
			    } 
			}
			SEMI
		) *
		RBRACE
	;

arg_sorts[boolean checkSort] returns [Sort[] argSorts = null] 
{
	Vector args = new Vector();
	Sort s;
}
	:
		(
			LPAREN
			s = sort_id_check[checkSort] { args.add(s); }
			(
				COMMA s = sort_id_check[checkSort] {args.add(s);}
			) *
			RPAREN
		) ?
		{
			argSorts = (Sort[])args.toArray(AN_ARRAY_OF_SORTS);
		}
	;

heur_decls :
		"heuristics"
		LBRACE
		(
			id:IDENT SEMI
			{ 
			    if (!skip_heuristics) {
				Heuristic h = new Heuristic(new Name(id.getText()));
				heur_ns.add(h);
			    }
			}
		) *
		RBRACE
	;

sort_id  returns [Sort s = null]
	:
		s = sort_id_check[true]
	;			

sort_id_check [boolean checkSort] returns [Sort s = null]
	:
		id:IDENT
                { 	String idtext=(useInstead.get(id.getText()));
			s= (Sort) sort_ns.lookup(new Name(idtext));
			if (checkSort && s == null) {
			    throw 
				new NotDeclException("sort", id, 
						     getFilename());
			}
		}
	;

id returns [String name = null]
	:
		classID:IDENT {
			name = classID.getText();
		} 
		( COLON COLON nameID:IDENT 
			{ 
			 name = classID.getText() + 
			        (nameID == null ? "" : 
			        "::" + nameID.getText());
			}
		)?	
		 
		
	;

