// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

/* -*-antlr-*- */
header {

    package de.uka.ilkd.key.parser;

    import java.io.*;
    import antlr.*;
    import de.uka.ilkd.key.logic.*;
    import de.uka.ilkd.key.logic.sort.*;    
    import de.uka.ilkd.key.logic.op.*;
    import de.uka.ilkd.key.rule.metaconstruct.*;
    import de.uka.ilkd.key.rule.*;
    import java.util.Vector;
    import java.util.HashMap;
    import java.util.HashSet;
    import java.util.Iterator;
    import de.uka.ilkd.key.java.visitor.ProgramVariableCollector;
    import de.uka.ilkd.key.java.Recoder2KeY;
    import de.uka.ilkd.key.java.JavaReader;
    import de.uka.ilkd.key.java.JavaInfo;
    import de.uka.ilkd.key.java.Services;
    import de.uka.ilkd.key.proof.UseInsteadMap;
}

/** ONLY FOR TEST CASES.
 * A parser for first-order terms and formulae.  Variables in 
 * quantifiers are expected to be declared globally in the variable
 * namespace.  This parser is used for test cases, where you want to know
 * in advance which objects will represent bound variables.
 */  

class GlobalDeclarationTermParser extends TermParser;
options {
	importVocab=KeYLexer;
	k = 2;
	defaultErrorHandler=true;
}


{
    static final Sort[] AN_ARRAY_OF_SORTS = new Sort[0];
    static final Term[] AN_ARRAY_OF_TERMS = new Term[0];

    protected Namespace var_ns;
    protected Namespace func_ns;
    protected Namespace sort_ns;
    protected Namespace heur_ns;

    // these variables are set if a file is read in step by
    // step. This used when reading in LDTs because of cyclic
    // dependencies.
    protected boolean skip_functions;
    protected boolean skip_predicates;
    protected boolean skip_sorts;
    protected boolean skip_schemavariables;
    protected boolean skip_heuristics;
    protected UseInsteadMap useInstead=new UseInsteadMap();
    protected JavaReader javaReader;
    protected TermFactory tf;

    /** collects errors and warnings that occured during parsing */
    protected ErrorHandler errHandler = new ErrorHandler();

    public boolean errors = false;



    public GlobalDeclarationTermParser(TokenStream lexer,
				       TermFactory tf,
		  		       JavaReader jr,
				       Namespace var_ns,
				       Namespace func_ns,
				       Namespace sort_ns) {
	this(lexer,2);
	this.tf = tf;
	this.javaReader = jr;
	this.var_ns     = var_ns;
	this.func_ns    = func_ns;
	this.sort_ns    = sort_ns;
    }

    public GlobalDeclarationTermParser(TokenStream lexer,
				       TermFactory tf,
		  		       JavaInfo ji,
				       Namespace var_ns,
				       Namespace func_ns,
				       Namespace sort_ns) {
	this(lexer, TermFactory.DEFAULT, 
		new Recoder2KeY(ji.rec2key(), ji.getTypeConverter()),
   	        var_ns, func_ns, sort_ns);	
    }

    public GlobalDeclarationTermParser(TokenStream lexer,
				       JavaInfo ji,
				       Namespace var_ns,
				       Namespace func_ns,
				       Namespace sort_ns) {
	this(lexer, TermFactory.DEFAULT,
		new Recoder2KeY(ji.rec2key(), ji.getTypeConverter()),
	        var_ns, func_ns, sort_ns);
    }

    protected int getLine() {
	int line = -1;
	try {
	    line = LT(0).getLine();
	} 
	catch (Exception e) {
	    System.err.println("No further token in stream");
	}
	return line;
    }   

    protected int getColumn() {
	int col = -1;
	try {
	    col = LT(0).getColumn();
	} 
	catch (Exception e) {
	    System.err.println("No further token in stream");
	}
	return col;
    }   

    // These three methods are required because the unused productions
    // from terms.g are inherited.  They should not be called.
    protected LogicVariable bindVar(Token id,Sort s) {
	throw new RuntimeException("bindVar was called in " +
				   "GlobalDeclarationTermParser.");
    }
	
    protected ProgramVariable progVar(Token id,Sort s) {
	var_ns.lookup(new Name(id.getText()));
	ProgramVariable v = (ProgramVariable)(ProgramVariable)
	    var_ns.lookup(new Name(id.getText()));
	if (v == null) {
	    v = new ProgramVariable(new ProgramElementName(id.getText()), s);
	    var_ns=var_ns.extended(v);
	}
	return v;
    }

    protected void unbindVars() {
	throw new RuntimeException("unbindVars was called in " +
				   "GlobalDeclarationTermParser.");
    }


    protected static HashSet progVars(JavaBlock jb) {
	ProgramVariableCollector pvc
	    = new ProgramVariableCollector(jb.program());
	pvc.start();
	return pvc.result();
    }


    protected void bindProgVars(HashSet progVars) {
	var_ns=new Namespace(var_ns);
	Iterator it=progVars.iterator();
	while (it.hasNext()) {
	    var_ns.add((ProgramVariable)it.next());
	}
    }

    protected void unbindProgVars() {
	var_ns = var_ns.parent();
    }


    protected Term termForParsedVariable(ParsableVariable v) 
	throws antlr.SemanticException 
	{
	    if ( v instanceof LogicVariable ) {
		return tf.createVariableTerm((LogicVariable)v);
	    } else {
		throw new antlr.SemanticException(
						  ""+ v +" is not a logic variable");
	    }
	}

    protected JavaBlock getJavaBlock(Token t) {
	return javaReader.readBlockWithEmptyContext(t.getText());
    }

    /** returns the java information object */
    public JavaInfo getJavaInfo() {
	return javaReader.createJavaInfo(new Services());
    }


	

}


top {Term t;} : t=formula;

bindingterm returns [Term a = null]
	:
		a = global_bindingterm
	;

/** here is the special behaviour, that requires a bound variable
 * to be previously declared in <code>var_ns</code>. */

global_bindingterm returns [Term a = null]
{
	Operator op = null;
	ParsableVariable v0, v1;
	LogicVariable v = null;
	Term a1,a2;
}
	:
		( 
			"eps" { op = Op.EPSILON; }
		|	"all" { op = Op.ALL; }
		|   "ex"  { op = Op.EX;  }
		)
		
		v0=var_id DOT // Don't need the sort here
		{v = (LogicVariable) v0;} a1=term60 
		{ 
			if ( op == Op.EPSILON ) {
				a = tf.createEpsilonTerm(v,a1);
			} else {
				a = tf.createQuantifierTerm((Quantifier)op, v,a1);
			}
		}
	|
		LBRACE v1=var_id a1=term RBRACE // Don't need the sort here
		{v = (LogicVariable) v1;} a2=term60
		{ 
			a = tf.createSubstitutionTerm(Op.SUBST,v,a1,a2);
		}
	;
