// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.proof;

import javax.swing.DefaultComboBoxModel;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.java.JavaInfo;
import java.io.*;
import de.uka.ilkd.key.rule.RuleApp;
import de.uka.ilkd.key.rule.TacletApp;
import de.uka.ilkd.key.rule.SyntacticalReplaceVisitor;
import de.uka.ilkd.key.rule.inst.*;


public class IfChoiceModel extends DefaultComboBoxModel {


    private static final String manualText="Manual Input";
    private String manualInput;
    private RuleApp app;
    private Term ifFma;

   /** namespace of variables */
    private Namespace var_ns;
    /** namespace of functions */
    private Namespace func_ns;
    /** namespace of sorts */
    private Namespace sort_ns;
    /** the java information object */
    private JavaInfo javaInfo;


    public IfChoiceModel(Term ifFma, TacletApp app, Sequent seq, 
			 JavaInfo javaInfo,
			 Namespace var_ns, Namespace func_ns, 
			 Namespace sort_ns) {
	super(ifInstsTerm(ifFma, app, seq, javaInfo));

	this.ifFma    = ifFma;
	this.app      = app;
	this.javaInfo = javaInfo;
	this.var_ns   = var_ns;
	this.func_ns  = func_ns;
	this.sort_ns  = sort_ns;

	addElement(manualText);
	manualInput="";
    }

    public String manualText() {
	return manualText;
    }

    public void setInput(String s) {
	manualInput=s;
    }

    public Term ifFma() {
	return ifFma;
    }
    

    public static Term[] ifInstsTerm(Term ifFma, TacletApp app, Sequent seq,
				     JavaInfo javaInfo) {	
 	ListOfSVInstantiations ifInsts
 	    =(app.taclet()).matchIf(seq, app.instantiations(), 
				  javaInfo);
	Term[] result = new Term[ifInsts.size()];
	IteratorOfSVInstantiations instsIt=ifInsts.iterator();
	for  (int j=0; j<result.length; j++) {
 		SyntacticalReplaceVisitor srv
 		    =new SyntacticalReplaceVisitor(javaInfo, instsIt.next()); 
 		ifFma.execPostOrder(srv);
		result[j]=srv.getTerm();
	}
	return result;
    }

    public static String[] ifInsts(Term ifFma, TacletApp app, Sequent seq, 
				   JavaInfo javaInfo) {	
	Term[] ifTerms = ifInstsTerm(ifFma, app, seq, javaInfo);
	String[] result = new String[ifTerms.length];	
	for (int i=0; i<result.length; i++) {
	    result[i] = ifTerms[i].toString();
	}
	return result;
    }



    /** creates a new Termparser that parses the given string
     * @param s the String to parse 
     */
    private TermParser stringParser(String s) {
	return new TermParser(new KeYLexer(new StringReader(s)),
			      "", TermFactory.DEFAULT, javaInfo,
			      var_ns, func_ns, sort_ns);
    }


    /** parses the given string that represents a term
     * @param s the String to parse 
     */
    public Term parseFormula(String s) throws antlr.ANTLRException {
	TermParser p = stringParser(s);
	return p.formula();
    }


    public Term getSelection() throws SVInstantiationException {
	Object o = getSelectedItem();
	if (o != manualText) {
	    return (Term)o;
	}
	try {
 	    if (manualInput == null || manualInput.equals("")) {
		return null;
	    }
	    return parseFormula(manualInput);
	} catch (antlr.ANTLRException e) {
	    throw new SVInstantiationException
		("Manual input for Taclet 'if "+ifFma+"' could not be parsed.", "",
		 -1, -1, e);
	} catch (Exception ex) {
	    throw new SVInstantiationException
		("Problems with parsing of Taclet if '"+ifFma+"'.", "", -1, -1, ex);
	}
    }

}
