// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.proof;

import antlr.*;

public class SVInstantiationException extends Exception {

    private String descr;        
    private String inst;
    private int row;
    private int column;
    private Exception exc;
         
    public SVInstantiationException(String descr, String inst, 
				  int row, int column,
				  Exception exc) {
	super("Instantiation Error");
	this.descr = descr;
	this.inst = inst;
	this.row = row;
	this.column = column;
	this.exc = exc;
    }


    public Exception getOriginalException() {
	return exc;
    }
    
    public int getRow() {
	return row;
    }

    public int getColumn() {
	return column;
    }

    private String space(int i) {
	StringBuffer res=new StringBuffer();
	for (int j=0; j<i; j++) {
	    res.append(" ");
	}
	return res.toString();
    }
    
    public String getMessage ()
    {
	String errmsg = descr;
	errmsg += row    <= 0 ? "" : ("\nRow: " + row);
	errmsg += column <= 0 ? "" : ("\nColumn: " + column);	
	//needs non-prop font:	errmsg +="\n"+inst;
 	if (column>0) {
	    //needs non-prop font: errmsg +="\n"+space(column-1)+"^";
	    StringBuffer sb=new StringBuffer(inst);
	    sb.insert(column-1, "\n"+space(column-1));
	    errmsg +="\n"+sb.toString();
 	} else errmsg+=inst;
	
	errmsg += (getOriginalException() == null ? "" : 
		   "\n"+getOriginalException().getMessage());
	return errmsg;
    }    

    /**
     * Returns a string representation of this exception.
     */
    public String toString() {
	return getMessage();
    }
}
