// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001 Universitaet Karlsruhe, Germany
//                    and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General 
// Public License. See LICENSE.TXT for details.
//
/**
 * tests the symbolic execution of the program meta constructs
 */
package de.uka.ilkd.key.rule.metaconstruct;

import junit.framework.*;
import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.expression.*;
import de.uka.ilkd.key.java.reference.*;
import de.uka.ilkd.key.java.statement.*;
import de.uka.ilkd.key.java.expression.literal.BooleanLiteral;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.rule.inst.*;
import de.uka.ilkd.key.java.visitor.*;
import java.io.*;

public class TestProgramMetaConstructs extends TestCase {

    private Recoder2KeY r2k;

    public TestProgramMetaConstructs(String name) {
	super(name);
	de.uka.ilkd.key.util.Debug.ENABLE_DEBUG = false;
    }

    public void setUp() {
	JavaInfo ji=new JavaInfo();
	r2k = new Recoder2KeY("/usr/lib/java/jre/lib/rt.jar", 
			      ji.rec2key(), ji.getTypeConverter());
    }

        
    public void testDoBreak() {
	LabeledStatement labeledBlock = (LabeledStatement)
	    ((StatementBlock)TacletForTests.parsePrg
	 //("{l1:l2:{l3:{l4:break l3; int i = 0;} int j=1;}}")).getChildAt(0);
	     ("{l4:break l3; int i = 0; int j=1;}")).getChildAt(0);
	DoBreak rmLabel = 
	    new DoBreak(labeledBlock);
	
	ProgramElement result = rmLabel.
	    symbolicExecution
	    (rmLabel.body(), new JavaInfo(),
	     SVInstantiations.EMPTY_SVINSTANTIATIONS);
	assert(result instanceof Break);
    }


    /** tests  AST walkers */
    public void xtestASTWalker() {
 	ProgramElement block = 
 	    TacletForTests.parsePrg("{int a=5; test1:test2:while (true) "+
 			  "{test3: {int j=3;}}}");
 	JavaASTCollector coll = new JavaASTCollector
 	    (block, 
 	     de.uka.ilkd.key.java.statement.LabeledStatement.class);
 	coll.start();
 	assert(coll.getNodes().size() == 3);

 	ProgramElement block2 = 
 	    TacletForTests.parsePrg("{while(true) {if (true) break; else continue;}}");
 	WhileLoopTransformation trans = 
 	    new WhileLoopTransformation(block2, 
					new ProgramElementName("l1"), 
					new ProgramElementName("l2"));
 	trans.start();
 	System.out.println("Result:"+trans);	
    }

    /** tests empty-try-to-if */
    public void testEmptyTryToIf() {
 	StatementBlock block = (StatementBlock)
 	    TacletForTests.parsePrg("{try{} catch(Exception e) {} finally{int a = 1;} }");
	ProgramMetaConstruct emptyTry = new EmptyTryToIf(block.getStatementAt(0));
	//	System.out.println(emptyTry.symbolicExecution(block.getStatementAt(0), null));

    }

    /** tests throw-try-to-if */
    public void testTryThrowToIf() {
 	StatementBlock block = (StatementBlock)
 	    TacletForTests.parsePrg("{Exception e; try{throw e;} catch(Exception e1) {int c = 0;} catch(RuntimeException e2) {int b = 0;} finally{int a = 1;} }");
	ProgramMetaConstruct tryThrow = new TryThrowToIf(block.getStatementAt(1));
	//	System.out.println(tryThrow.symbolicExecution(block.getStatementAt(1), null));
 	StatementBlock block2 = (StatementBlock)
 	    TacletForTests.parsePrg("{Exception e; try{throw e;} catch(Exception e1) {int c = 0;} catch(RuntimeException e2) {int b = 0;}  }");
	ProgramMetaConstruct tryThrow2 = new TryThrowToIf(block2.getStatementAt(1));
	//	System.out.println(tryThrow2.symbolicExecution(block2.getStatementAt(1), null));

    }

    /** tests instance-of */
    public void testInstanceof() {
	JavaInfo ji=TacletForTests.r2k.createJavaInfo(new Services());
 	StatementBlock block = (StatementBlock)
 	    TacletForTests.parsePrg("{ boolean b; String e;"
				  +" b = e instanceof String; }");
	ProgramMetaConstruct instof 
	    = new InstanceOfEval
		((Expression)((Assignment)block.getStatementAt(2))
		 .getChildAt(1));
	assert(((BooleanLiteral)instof.symbolicExecution
		(((Assignment)block.getStatementAt(2)).getChildAt(1), ji
		 , 
		 SVInstantiations.EMPTY_SVINSTANTIATIONS)).getValue());
	block = (StatementBlock)
 	    TacletForTests.parsePrg("{ boolean b; String e;"
				  +" b = e instanceof Integer; }");
	instof = new InstanceOfEval((Expression)
				    ((Assignment)block.getStatementAt(2))
				    .getChildAt(1));
	assert(!((BooleanLiteral)instof.symbolicExecution
		 (((Assignment)block.getStatementAt(2)).getChildAt(1), ji, 
		  SVInstantiations.EMPTY_SVINSTANTIATIONS)).getValue());
    }

    public void testTypeOf() { //this is no really sufficient test
	JavaInfo ji=TacletForTests.r2k.createJavaInfo(new Services());
	//but I can't access programs here
 	StatementBlock block = (StatementBlock)
 	    TacletForTests.parsePrg(" { int i; int j; i=j; }");
	Expression expr=(Expression)((Assignment)block.getStatementAt(2))
	    .getChildAt(1);
	ProgramMetaConstruct typeof = new TypeOf(expr);
	assert(((TypeRef)typeof.symbolicExecution(expr, ji, 
		 SVInstantiations.EMPTY_SVINSTANTIATIONS)).getName()
	       .toString().equals("int"));
    }

}
