// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.util.install;

import java.util.Vector;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;


public class Installer extends JFrame {


    public static int RUNNING=-1;
    public static int CANCEL=0;
    public static int FINISH=1;

    private JTabbedPane tabbedPane;
    private int exitCode=RUNNING;
    
    public Installer(String title) {
	super(title);
	tabbedPane=new JTabbedPane();

	getContentPane().setLayout(new BorderLayout());
	getContentPane().add(tabbedPane, BorderLayout.CENTER);
	getContentPane().add(setDefaultButtonPanel(), 
			     BorderLayout.SOUTH);		
    }

    public void setHeader(JLabel header) {
	getContentPane().add(header, BorderLayout.NORTH);
    }

    public void addInstallationPane(InstallationPane pane) {
	tabbedPane.addTab(pane.name(), pane);
    }

    public int paneCount() {
	return tabbedPane.getTabCount();
    }

    public InstallationPane pane(int index) {
	return (InstallationPane)tabbedPane.getComponentAt(index);
    }

    public int exitCode() {
	return exitCode;
    }

    protected Container setDefaultButtonPanel() {
	JPanel buttonPanel=new JPanel();
	buttonPanel.setAlignmentX(SwingConstants.CENTER);	
	Box buttonBox=Box.createHorizontalBox();
	
	// create buttons
	JButton prev=new JButton("<< Back");
	JButton cancel=new JButton("Cancel");
	JButton finish=new JButton("Finish");
	JButton next=new JButton("Next >>");
	// add to change listeners
	ChangeListenerWithButton prevListener=new ChangeListenerWithButton(prev) {
		private void defaultAction() {
		    if (tabbedPane.getSelectedIndex()<=0) {		    
			button.setEnabled(false);
		    } else {
			button.setEnabled(true);
		    }
		}

		public void stateChanged(ChangeEvent ce) {
		    defaultAction();
		}

		public void componentAdded(ContainerEvent e) {
		    defaultAction();
		}
		public void componentRemoved(ContainerEvent e) {
		    defaultAction();
		}		
	    };
	tabbedPane.addContainerListener(prevListener);
	tabbedPane.addChangeListener(prevListener); 
	
	ChangeListenerWithButton nextListener=new ChangeListenerWithButton(next) {
		private void defaultAction() {
		    if (tabbedPane.getSelectedIndex()==tabbedPane.getTabCount()-1) {		    
			button.setEnabled(false);
		    } else {
			button.setEnabled(true);
		    }
		}
		
		public void stateChanged(ChangeEvent ce) {
		    defaultAction();
		}
		
		public void componentAdded(ContainerEvent e) {
		    defaultAction();
		}
		public void componentRemoved(ContainerEvent e) {
		    defaultAction();
		}		
	    };
	tabbedPane.addContainerListener(nextListener); 
	tabbedPane.addChangeListener(nextListener); 


	// set action listeners
	prev.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    if (tabbedPane.getSelectedIndex()>0) {
			tabbedPane.setSelectedIndex(tabbedPane.getSelectedIndex()-1);
		    }
		}
	    });
	cancel.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    exitCode=Installer.CANCEL;
		    setVisible(false);
		    dispose();
		}
	    });
	finish.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    exitCode=Installer.FINISH;
		    setVisible(false);
		    dispose();
		}
	    });
	next.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    if (tabbedPane.getSelectedIndex()<tabbedPane.getTabCount()-1) {
			tabbedPane.setSelectedIndex(tabbedPane.getSelectedIndex()+1);
		    }
		}
	    });
	
	// add buttons
	buttonBox.add(prev);
	buttonBox.add(cancel);
	buttonBox.add(finish);
	buttonBox.add(next);
	buttonPanel.add(buttonBox);
	return buttonPanel;
     }

    // Change listener for tabbed pane
    static abstract class ChangeListenerWithButton 
	implements ChangeListener, ContainerListener {

	protected JButton button;

	public ChangeListenerWithButton(JButton button) {
	    this.button=button;
	}

	public abstract void stateChanged(ChangeEvent ce);

	public abstract void componentAdded(ContainerEvent e);
	public abstract void componentRemoved(ContainerEvent e);
  

    }
}
